/*******************************************************************************
 * Copyright (c) PLX Technology, Inc.
 *
 * PLX Technology Inc. licenses this source file under the GNU Lesser General Public
 * License (LGPL) version 2.  This source file may be modified or redistributed
 * under the terms of the LGPL and without express permission from PLX Technology.
 *
 * PLX Technology, Inc. provides this software AS IS, WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING, WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  PLX makes no guarantee
 * or representations regarding the use of, or the results of the use of,
 * the software and documentation in terms of correctness, accuracy,
 * reliability, currentness, or otherwise; and you rely on the software,
 * documentation and results solely at your own risk.
 *
 * IN NO EVENT SHALL PLX BE LIABLE FOR ANY LOSS OF USE, LOSS OF BUSINESS,
 * LOSS OF PROFITS, INDIRECT, INCIDENTAL, SPECIAL OR CONSEQUENTIAL DAMAGES
 * OF ANY KIND.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * File Name:
 *
 *      PlxChipFn.c
 *
 * Description:
 *
 *      Contains PLX chip-specific support functions
 *
 * Revision History:
 *
 *      06-01-08 : PLX SDK v6.00
 *
 ******************************************************************************/


#include "ApiFunc.h"
#include "PciFunc.h"
#include "PlxChipFn.h"
#include "PlxInterrupt.h"
#include "SuppFunc.h"




/******************************************************************************
 *
 * Function   :  PlxChipSetInterruptRegisterOffsets
 *
 * Description:  Determines the IRQ register offsets, based on chip type, for the ISR
 *
 *****************************************************************************/
BOOLEAN
PlxChipSetInterruptRegisterOffsets(
    DEVICE_EXTENSION *pdx
    )
{
    U32 OffsetIrqBase;



    /*****************************************************************
     * The registers used are the IRQ Set Mask and IRQ Clear Mask.
     * Calculated offsets are based on the following:
     *
     *    Virtual-side IRQ Set         -- Base + 00h
     *    Virtual-side IRQ Clear       -- Base + 04h
     *    Virtual-side IRQ Set Mask    -- Base + 08h
     *    Virtual-side IRQ Clear Mask  -- Base + 0Ch
     *    Link-side IRQ Set            -- Base + 10h
     *    Link-side IRQ Clear          -- Base + 14h
     *    Link-side IRQ Set Mask       -- Base + 18h
     *    Link-side IRQ Clear Mask     -- Base + 1Ch
     *
     ****************************************************************/

    // Add offset based on chip type
    switch (pdx->PlxChipType & 0xFF00)
    {
        case 0x8500:
            OffsetIrqBase = 0x90;
            break;

        case 0x8600:
            OffsetIrqBase = 0xC4C;
            break;

        case 0:
        default:
            ErrorPrintf(("Error: Invalid PLX chip, unable to determine IRQ register offsets\n"));
            return FALSE;
    }

    // Add offset based on virtual or link side
    if (pdx->bFromLinkSide)
    {
        // Add link-side base
        OffsetIrqBase += 0x11000;

        // Skip over Virtual-side registers
        OffsetIrqBase += 0x10;
    }
    else
    {
        // Add Virtual-side base
        OffsetIrqBase += 0x10000;
    }

    // Set final offsets
    pdx->Offset_DB_IntStatus    = OffsetIrqBase + 0x00;
    pdx->Offset_DB_IntClear     = OffsetIrqBase + 0x04;
    pdx->Offset_DB_IntMaskSet   = OffsetIrqBase + 0x08;
    pdx->Offset_DB_IntMaskClear = OffsetIrqBase + 0x0C;

    return TRUE;
}




/******************************************************************************
 *
 * Function   :  PlxChipInterruptsEnable
 *
 * Description:  Globally enables PLX chip interrupts
 *
 *****************************************************************************/
BOOLEAN
PlxChipInterruptsEnable(
    DEVICE_EXTENSION *pdx
    )
{
    // Enable doorbell interrupts
    PLX_8000_REG_WRITE(
        pdx,
        pdx->Offset_DB_IntMaskClear,
        0xFFFF
        );

    return TRUE;
}




/******************************************************************************
 *
 * Function   :  PlxChipInterruptsDisable
 *
 * Description:  Globally disables PLX chip interrupts
 *
 *****************************************************************************/
BOOLEAN
PlxChipInterruptsDisable(
    DEVICE_EXTENSION *pdx
    )
{
    // Disable doorbell interrupts
    PLX_8000_REG_WRITE(
        pdx,
        pdx->Offset_DB_IntMaskSet,
        0xFFFF
        );

    return TRUE;
}




/******************************************************************************
 *
 * Function   :  PlxDetermineNtPortSide
 *
 * Description:  Determines whether the NT port is Virtual or Link side
 *
 ******************************************************************************/
BOOLEAN
PlxDetermineNtPortSide(
    DEVICE_EXTENSION *pdx
    )
{
    U32 RegValue;


    // Default to virtual side
    pdx->bFromLinkSide = FALSE;

    // Check if BAR 0 exists
    if (pdx->PciBar[0].Properties.Physical == 0)
    {
        // If BAR 0 not enabled, this is 8500 virtual-side
    }
    else
    {
        if (pdx->pRegVa == NULL)
        {
            DebugPrintf(("Error: Unable to determine NT side - BAR not mapped for reg access\n"));
            return FALSE;
        }

        // Read internal link-side BAR 0
        RegValue =
            PLX_8000_REG_READ(
                pdx,
                (0x11000 + 0x10)     // NT Link-side BAR 0
                );

        if (RegValue == 0)
        {
            DebugPrintf(("Error: Unable to determine NT side - Internal NT BAR 0 is 0\n"));
            return FALSE;
        }

        // If internal BAR 0 matches PCI BAR 0, port is link side
        if (RegValue == pdx->PciBar[0].Properties.BarValue)
            pdx->bFromLinkSide = TRUE;
    }

    DebugPrintf((
        "NT port is %s-side\n",
        (pdx->bFromLinkSide == TRUE) ? "Link" : "Virtual"
        ));

    return TRUE;
}




/******************************************************************************
 *
 * Function   :  PlxMapRegisterBar
 *
 * Description:  Maps the correct BAR space for register accesses
 *
 ******************************************************************************/
BOOLEAN
PlxMapRegisterBar(
    DEVICE_EXTENSION *pdx
    )
{
    U32 PciBar;
    U32 BarSize;


    if (pdx->PciBar[0].pVa != NULL)
    {
        // Use PCI BAR 0 kernel address
        pdx->pRegVa = (U8*)(pdx->PciBar[0].pVa);

        if (pdx->pRegVa == NULL)
        {
            return FALSE;
        }

        DebugPrintf((
            "Using PCI BAR 0 (VA=%p) ==> PLX regs\n",
            pdx->pRegVa
            ));
    }
    else
    {
        // Get the BAR 0 value from the upstream port
        PlxPciRegisterRead(
            pdx->Key.bus - 1,   // Bus will be our bus minus 1
            0,                  // Always at slot 0
            0,
            0x10,               // BAR 0
            &PciBar
            );

        if (PciBar == 0)
        {
            return FALSE;
        }

        // Query BAR range
        PlxPciRegisterWrite(
            pdx->Key.bus - 1,
            0,
            0,
            0x10,               // BAR 0
            (U32)-1
            );

        // Read size
        PlxPciRegisterRead(
            pdx->Key.bus - 1,
            0,
            0,
            0x10,               // BAR 0
            &BarSize
            );

        // Restore BAR
        PlxPciRegisterWrite(
            pdx->Key.bus - 1,
            0,
            0,
            0x10,               // BAR 0
            PciBar
            );

        // Calculate BAR size
        BarSize = (~(BarSize & ~0xF)) + 1;

        // Clear any non-address bits
        PciBar &= ~(0xF);

        // Save BAR size
        pdx->UpstreamBarSize = BarSize;

        // Map the space
        pdx->pRegVa =
            ioremap(
                PciBar,
                BarSize
                );

        if (pdx->pRegVa == NULL)
        {
            return FALSE;
        }

        DebugPrintf((
            "Mapped upstream BAR 0 (%08lX) ==> PLX regs (VA=%p) (%dkb)\n",
            (PLX_UINT_PTR)PciBar, pdx->pRegVa, (BarSize >> 10)
            ));
    }

    return TRUE;
}




/*******************************************************************************
 *
 * Function   :  PlxErrataWorkAround_NtBarShadow
 *
 * Description:  Implements work-around for NT BAR shadow errata
 *
 ******************************************************************************/
VOID
PlxErrataWorkAround_NtBarShadow(
    DEVICE_EXTENSION *pdx
    )
{
    U8  i;
    U16 offset;
    U32 RegValue;


    /*****************************************************
     * This work-around handles the errata for NT Virtual
     * side BAR registers.  For BARs 2-5, the BAR space
     * is not accessible until the shadow registers are
     * updated manually.
     *
     * The procedure is to first read the BAR Setup registers
     * and write them back to themselves.  Then the PCI BAR
     * register is read and the same value is written back.
     * This updates the internal shadow registers of the PLX
     * chip and makes the BAR space accessible for NT data
     * transfers.
     *
     * Note: The BAR setup register must be written before
     *       the PCI BAR register.
     ****************************************************/

    // If link side, not errata so return
    if (pdx->bFromLinkSide)
        return;

    DebugPrintf(("Implementing NT Virtual-side BAR shadow errata work-around...\n"));

    // Read & write-back BAR setup and BAR registers for BARs 2-5
    for (i = 2; i <= 5; i++)
    {
        // Set offset for NT BAR setup register (D4h = Virtual-side BAR 2 Setup)
        offset = 0xD4 + ((i-2) * sizeof(U32));

        // Get Virtual-side BAR setup register
        PLX_PCI_REG_READ(
            pdx,
            offset,
            &RegValue
            );

        // Check if BAR is enabled
        if (RegValue & (1 << 31))
        {
            // Write BAR Setup back to itself
            PLX_PCI_REG_WRITE(
                pdx,
                offset,
                RegValue
                );

            // Read the corresponding PCI BAR
            PLX_PCI_REG_READ(
                pdx,
                0x10 + (i * sizeof(U32)),
                &RegValue
                );

            // Write PCI BAR back to itself
            PLX_PCI_REG_WRITE(
                pdx,
                0x10 + (i * sizeof(U32)),
                RegValue
                );
        }
    }
}




/*******************************************************************************
 *
 * Function   :  PlxErrataWorkAround_NtCapturedRequesterID
 *
 * Description:  Implements work-around for NT Link-side captured requester ID
 *
 ******************************************************************************/
VOID
PlxErrataWorkAround_NtCapturedRequesterID(
    DEVICE_EXTENSION *pdx
    )
{
    U8            NonNtPort;
    PLX_PORT_PROP PortProp;


    // If virtual side, not errata so return
    if (pdx->bFromLinkSide == FALSE)
        return;

    DebugPrintf(("Implementing NT Link-side captured Requester ID work-around...\n"));

    // Get the port number to determine which Non-NT port to update
    PlxGetPortProperties(
        pdx,
        &pdx->Key,
        &PortProp
        );

    if ((U16)PortProp.PortNumber <= 3)
    {
        NonNtPort = 8;
    }
    else
    {
        NonNtPort = 0;
    }

    // Update captured bus/device number
    PlxRegisterWrite(
        pdx,
        &pdx->Key,
        (NonNtPort * (0x1000)) + 0xDF4,
        ((pdx->Key.bus << 0) & 0xFF) | ((pdx->Key.slot << 8) & 0x1F),
        FALSE          // Don't adjust offset for port
        );

    DebugPrintf((
        "Updated Non-NT station (port %d) with captured Requester ID [b:%02x s:%02x f:00]\n",
        NonNtPort, pdx->Key.bus, pdx->Key.slot
        ));
}
