/*******************************************************************************
 * Copyright (c) PLX Technology, Inc.
 *
 * PLX Technology Inc. licenses this source file under the GNU Lesser General Public
 * License (LGPL) version 2.  This source file may be modified or redistributed
 * under the terms of the LGPL and without express permission from PLX Technology.
 *
 * PLX Technology, Inc. provides this software AS IS, WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING, WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  PLX makes no guarantee
 * or representations regarding the use of, or the results of the use of,
 * the software and documentation in terms of correctness, accuracy,
 * reliability, currentness, or otherwise; and you rely on the software,
 * documentation and results solely at your own risk.
 *
 * IN NO EVENT SHALL PLX BE LIABLE FOR ANY LOSS OF USE, LOSS OF BUSINESS,
 * LOSS OF PROFITS, INDIRECT, INCIDENTAL, SPECIAL OR CONSEQUENTIAL DAMAGES
 * OF ANY KIND.
 *
 ******************************************************************************/

/*******************************************************************************
 *
 * File Name:
 *
 *      ApiFunc.c
 *
 * Description:
 *
 *      Implements the PLX API functions
 *
 * Revision History:
 *
 *      07-01-08 : PLX SDK v6.00
 *
 ******************************************************************************/


#include "ApiFunc.h"
#include "ChipFunc.h"
#include "Eep_6000.h"
#include "Eep_8000.h"
#include "Eep_8111.h"
#include "PciFunc.h"
#include "SuppFunc.h"




/*******************************************************************************
 *
 * Function   :  PlxDeviceFind
 *
 * Description:  Search for a specific device using device key parameters
 *
 ******************************************************************************/
PLX_STATUS
PlxDeviceFind(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U8               *pDeviceNumber
    )
{
    U8                DeviceCount;
    U32               RegValue;
    BOOLEAN           bMatchId;
    BOOLEAN           bMatchLoc;
    PLX_DEVICE_NODE  *pNode;
    struct list_head *pEntry;


    DeviceCount = 0;

    pEntry = pdx->List_SupportedDevices.next;

    // Traverse list to find the desired list objects
    while (pEntry != &(pdx->List_SupportedDevices))
    {
        // Get the object
        pNode =
            list_entry(
                pEntry,
                PLX_DEVICE_NODE,
                ListEntry
                );

        // Assume successful match
        bMatchLoc = TRUE;
        bMatchId  = TRUE;

        //
        // Compare device key information
        //

        // Compare Bus number
        if (pKey->bus != (U8)PCI_FIELD_IGNORE)
        {
            if (pKey->bus != pNode->Key.bus)
            {
                bMatchLoc = FALSE;
            }
        }

        // Compare Slot number
        if (pKey->slot != (U8)PCI_FIELD_IGNORE)
        {
            if (pKey->slot != pNode->Key.slot)
            {
                bMatchLoc = FALSE;
            }
        }

        // Compare Function number
        if (pKey->function != (U8)PCI_FIELD_IGNORE)
        {
            if (pKey->function != pNode->Key.function)
            {
                bMatchLoc = FALSE;
            }
        }

        //
        // Compare device ID information
        //

        // Compare Vendor ID
        if (pKey->VendorId != (U16)PCI_FIELD_IGNORE)
        {
            if (pKey->VendorId != pNode->Key.VendorId)
            {
                bMatchId = FALSE;
            }
        }

        // Compare Device ID
        if (pKey->DeviceId != (U16)PCI_FIELD_IGNORE)
        {
            if (pKey->DeviceId != pNode->Key.DeviceId)
            {
                bMatchId = FALSE;
            }
        }

        // Subsystem ID only valid in PCI header 0 type
        if (pNode->PciHeaderType == 0)
        {
            // Get the Subsystem Device/Vendor ID
            if (pNode->Key.SubVendorId == 0)
            {
                PLX_PCI_REG_READ(
                    pNode,
                    0x2c,        // PCI Subsystem ID
                    &RegValue
                    );

                pNode->Key.SubVendorId = (U16)(RegValue & 0xffff);
                pNode->Key.SubDeviceId = (U16)(RegValue >> 16);
            }

            // Compare Subsystem Vendor ID
            if (pKey->SubVendorId != (U16)PCI_FIELD_IGNORE)
            {
                if (pKey->SubVendorId != pNode->Key.SubVendorId)
                {
                    bMatchId = FALSE;
                }
            }

            // Compare Subsystem Device ID
            if (pKey->SubDeviceId != (U16)PCI_FIELD_IGNORE)
            {
                if (pKey->SubDeviceId != pNode->Key.SubDeviceId)
                {
                    bMatchId = FALSE;
                }
            }
        }

        // Get the Revision ID if haven't yet
        if (pNode->Key.Revision == 0)
        {
            PLX_PCI_REG_READ(
                pNode,
                0x08,        // PCI Revision ID
                &RegValue
                );

            pNode->Key.Revision = (U8)(RegValue & 0xFF);
        }

        // Compare Revision
        if (pKey->Revision != (U8)PCI_FIELD_IGNORE)
        {
            if (pKey->Revision != pNode->Key.Revision)
            {
                bMatchId = FALSE;
            }
        }

        // Check if match on location and ID
        if (bMatchLoc && bMatchId)
        {
            // Match found, check if it is the desired device
            if (DeviceCount == *pDeviceNumber)
            {
                // Copy the device information
                *pKey = pNode->Key;

                DebugPrintf((
                    "Criteria matched device %04X_%04X [b:%02x s:%02x f:%02x]\n",
                    pNode->Key.DeviceId, pNode->Key.VendorId,
                    pNode->Key.bus, pNode->Key.slot, pNode->Key.function
                    ));

                return ApiSuccess;
            }

            // Increment device count
            DeviceCount++;
        }

        // Jump to next entry
        pEntry = pEntry->next;
    }

    // Return number of matched devices
    *pDeviceNumber = DeviceCount;

    DebugPrintf(("Criteria did not match any devices\n"));

    return ApiInvalidDeviceInfo;
}




/*******************************************************************************
 *
 * Function   :  PlxChipTypeGet
 *
 * Description:  Returns PLX chip type and revision
 *
 ******************************************************************************/
PLX_STATUS
PlxChipTypeGet(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U32              *pChipType,
    U8               *pRevision
    )
{
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        *pChipType = 0x0;
        *pRevision = 0x0;

        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    *pChipType = pNode->PlxChipType;
    *pRevision = pNode->PlxRevision;

    DebugPrintf((
        "Device %04X_%04X = %04X rev %02X\n",
        pKey->DeviceId, pKey->VendorId, *pChipType, *pRevision
        ));

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxChipTypeSet
 *
 * Description:  Sets the PLX chip type dynamically
 *
 ******************************************************************************/
PLX_STATUS
PlxChipTypeSet(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U32               ChipType,
    U8                Revision
    )
{
    U32              TempChip;
    BOOLEAN          bSetUptreamNode;
    PLX_PORT_PROP    PortProp;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Default to no upstream node
    bSetUptreamNode = FALSE;

    // Attempt auto-detection if requested
    if (ChipType == 0)
    {
        PlxChipTypeDetect(
            pNode
            );

        ChipType = pNode->PlxChipType;
    }

    // Generalize by device type
    if (((ChipType & 0xFF00) == 0x8500) ||
        ((ChipType & 0xFF00) == 0x8600))
    {
        TempChip = 0x8000;
    }
    else
    {
        TempChip = ChipType;
    }

    // Verify chip type
    switch (TempChip)
    {
        case 0:         // Used to clear chip type
        case 0x6140:
        case 0x6150:
        case 0x6152:
        case 0x6154:
        case 0x6156:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
        case 0x8111:
        case 0x8112:
            break;

        case 0x8114:
            // Flag to update upstream node
            bSetUptreamNode = TRUE;
            break;

        case 0x8000:
            // Get port properties to ensure upstream node
            PlxGetPortProperties(
                pdx,
                pKey,
                &PortProp
                );

            if (PortProp.PortType != PLX_PORT_UPSTREAM)
            {
                DebugPrintf(("ERROR - Chip type may only be changed on upstream port\n"));
                return ApiUnsupportedFunction;
            }

            // Flag to update upstream node
            bSetUptreamNode = TRUE;
            break;

        default:
            DebugPrintf((
                "ERROR - Invalid or unsupported chip type (%04X)\n",
                (U16)ChipType
                ));
            return ApiInvalidData;
    }

    // Set the new chip type
    pNode->PlxChipType = ChipType;

    // Check if we should update the revision or use the default
    if ((Revision == (U8)-1) || (Revision == 0))
    {
        // Attempt to detect Revision ID
        PlxChipRevisionDetect(
            pNode
            );
    }
    else
    {
        pNode->PlxRevision = Revision;
    }

    if (bSetUptreamNode)
    {
        DebugPrintf(("Set parent upstream node to itself for register access\n"));
        pNode->pUpstreamNode = pNode;
    }
    else
    {
        pNode->pUpstreamNode = NULL;
    }

    DebugPrintf((
        "Set device (%04X_%04X) type to %04X rev %02X\n",
        pKey->DeviceId, pKey->VendorId, pNode->PlxChipType, pNode->PlxRevision
        ));

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxGetPortProperties
 *
 * Description:  Returns the current port information and status
 *
 ******************************************************************************/
PLX_STATUS
PlxGetPortProperties(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    PLX_PORT_PROP    *pPortProp
    )
{
    U16              MaxSize;
    U16              Offset_PcieCap;
    U32              RegValue;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Set default properties
    RtlZeroMemory(pPortProp, sizeof(PLX_PORT_PROP));

    // Get the offset of the PCI Express capability 
    Offset_PcieCap =
        PlxGetExtendedCapabilityOffset(
            pNode,
            0x10       // CAP_ID_PCI_EXPRESS
            );

    if (Offset_PcieCap == 0)
    {
        DebugPrintf(("Device does not support PCI Express Capability\n"));

        // Mark device as non-PCIe
        pPortProp->bNonPcieDevice = TRUE;

        // Default to a legacy endpoint
        if (pNode->PciHeaderType == 0)
            pPortProp->PortType = PLX_PORT_LEGACY_ENDPOINT;
        else
            pPortProp->PortType = PLX_PORT_UNKNOWN;

        return ApiSuccess;
    }

    // Get PCIe Capability
    PLX_PCI_REG_READ(
        pNode,
        Offset_PcieCap,
        &RegValue
        );

    // Get port type
    pPortProp->PortType = (U8)((RegValue >> 20) & 0xF);

    // Get PCIe Device Control
    PLX_PCI_REG_READ(
        pNode,
        Offset_PcieCap + 0x08,
        &RegValue
        );

    // Get max payload size field
    MaxSize = (U8)(RegValue >> 5) & 0x7;

    // Set max payload size (=128 * (2 ^ MaxPaySizeField))
    if (MaxSize <= 5)
        pPortProp->MaxPayloadSize = 128 * ((U16)Plx_pow_int(2, MaxSize));

    // Get max read request size field
    MaxSize = (U8)(RegValue >> 12) & 0x7;

    // Set max read request size (=128 * (2 ^ MaxReadReqSizeField))
    if (MaxSize <= 5)
        pPortProp->MaxReadReqSize = 128 * ((U16)Plx_pow_int(2, MaxSize));

    // Get PCIe Link Capabilities
    PLX_PCI_REG_READ(
        pNode,
        Offset_PcieCap + 0x0C,
        &RegValue
        );

    // Get port number
    pPortProp->PortNumber = (U8)((RegValue >> 24) & 0xFF);

    // Get max link width
    pPortProp->MaxLinkWidth = (U8)((RegValue >> 4) & 0x3F);

    // Get max link speed
    pPortProp->MaxLinkSpeed = (U8)((RegValue >> 0) & 0xF);

    // Get PCIe Link Status/Control
    PLX_PCI_REG_READ(
        pNode,
        Offset_PcieCap + 0x10,
        &RegValue
        );

    // Get link width
    pPortProp->LinkWidth = (U8)((RegValue >> 20) & 0x3F);

    // Get link speed
    pPortProp->LinkSpeed = (U8)((RegValue >> 16) & 0xF);

    // Store the port number in the device properties
    pNode->PortNumber = pPortProp->PortNumber;

    DebugPrintf((
        "Type=%d Num=%d MaxPd=%d MaxRdReq=%d LW=x%d/x%d LS=%d/%d\n",
        pPortProp->PortType, pPortProp->PortNumber,
        pPortProp->MaxPayloadSize, pPortProp->MaxReadReqSize,
        pPortProp->LinkWidth, pPortProp->MaxLinkWidth,
        pPortProp->LinkSpeed, pPortProp->MaxLinkSpeed
        ));

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxPciDeviceReset
 *
 * Description:  Resets a device using software reset feature of PLX chip
 *
 ******************************************************************************/
PLX_STATUS
PlxPciDeviceReset(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey
    )
{
    // Device reset not implemented
    return ApiUnsupportedFunction;
}




/*******************************************************************************
 *
 * Function   :  PlxRegisterRead
 *
 * Description:  Reads a PLX-specific control register
 *
 ******************************************************************************/
U32
PlxRegisterRead(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U32               offset,
    PLX_STATUS       *pStatus,
    BOOLEAN           bAdjustForPort
    )
{
    U32              TempChip;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        if (pStatus != NULL)
            *pStatus = ApiInvalidDeviceInfo;
        return 0;
    }

    // Generalize by device type
    if (((pNode->PlxChipType & 0xFF00) == 0x8500) ||
        ((pNode->PlxChipType & 0xFF00) == 0x8600))
    {
        TempChip = 0x8000;
    }
    else
    {
        TempChip = pNode->PlxChipType;
    }

    // Verify device is supported
    switch (TempChip)
    {
        case 0x8111:
        case 0x8112:
            // Driver does not support 8111/8112 memory mapped register access
            if (bAdjustForPort)
            {
                return PlxRegisterRead_8111(
                    pNode,
                    offset,
                    pStatus
                    );
            }
            break;

        case 0x8114:
        case 0x8000:
            return PlxRegisterRead_8000(
                pNode,
                offset,
                pStatus,
                bAdjustForPort
                );
    }

    if (pStatus != NULL)
        *pStatus = ApiUnsupportedFunction;

    return 0;
}




/*******************************************************************************
 *
 * Function   :  PlxRegisterWrite
 *
 * Description:  Writes to a PLX-specific control register
 *
 ******************************************************************************/
PLX_STATUS
PlxRegisterWrite(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U32               offset,
    U32               value,
    BOOLEAN           bAdjustForPort
    )
{
    U32              TempChip;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Generalize by device type
    if (((pNode->PlxChipType & 0xFF00) == 0x8500) ||
        ((pNode->PlxChipType & 0xFF00) == 0x8600))
    {
        TempChip = 0x8000;
    }
    else
    {
        TempChip = pNode->PlxChipType;
    }

    // Verify device is supported
    switch (TempChip)
    {
        case 0x8111:
        case 0x8112:
            // Driver does not support 8111/8112 memory mapped register access
            if (bAdjustForPort)
            {
                return PlxRegisterWrite_8111(
                    pNode,
                    offset,
                    value
                    );
            }
            break;

        case 0x8114:
        case 0x8000:
            return PlxRegisterWrite_8000(
                pNode,
                offset,
                value,
                bAdjustForPort
                );
    }

    return ApiUnsupportedFunction;
}




/*******************************************************************************
 *
 * Function   :  PlxPciBarProperties
 *
 * Description:  Returns the properties of a PCI BAR space
 *
 ******************************************************************************/
PLX_STATUS
PlxPciBarProperties(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U8                BarIndex,
    PLX_PCI_BAR_PROP *pBarProperties
    )
{
    U32              PciBar;
    U32              Size;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Verify BAR index
    switch (pNode->PciHeaderType)
    {
        case 0:
            if ((BarIndex != 0) && (BarIndex > 5))
                return ApiInvalidIndex;
            break;

        case 1:
            if ((BarIndex != 0) && (BarIndex != 1))
                return ApiInvalidIndex;
            break;

        default:
            return ApiInvalidIndex;
    }

    // Get BAR properties if not probed yet
    if (pNode->PciBar[BarIndex].Properties.BarValue == 0)
    {
        // Read PCI BAR
        PLX_PCI_REG_READ(
            pNode,
            0x10 + (BarIndex * sizeof(U32)),
            &PciBar
            );

        // Query BAR range
        PLX_PCI_REG_WRITE(
            pNode,
            0x10 + (BarIndex * sizeof(U32)),
            (U32)-1
            );

        // Read size
        PLX_PCI_REG_READ(
            pNode,
            0x10 + (BarIndex * sizeof(U32)),
            &Size
            );

        // Restore BAR
        PLX_PCI_REG_WRITE(
            pNode,
            0x10 + (BarIndex * sizeof(U32)),
            PciBar
            );

        // If upper 32-bit of 64-bit BAR, all bits will be set, which is not supported
        if (Size == (U32)-1)
            Size = 0;

        // Store BAR properties
        pNode->PciBar[BarIndex].Properties.BarValue      = PciBar;
        pNode->PciBar[BarIndex].Properties.bIoSpace      = (BOOLEAN)((PciBar >> 0) & (1 << 0));
        pNode->PciBar[BarIndex].Properties.bPrefetchable = (BOOLEAN)((PciBar >> 3) & (1 << 0));
        pNode->PciBar[BarIndex].Properties.b64bit        = (BOOLEAN)((PciBar >> 2) & (1 << 0));

        // Determine BAR address & size based on space type
        if (pNode->PciBar[BarIndex].Properties.bIoSpace)
        {
            pNode->PciBar[BarIndex].Properties.Physical = PciBar & ~0x3;
            pNode->PciBar[BarIndex].Properties.Size     = (~(Size & ~0x3)) + 1;
        }
        else
        {
            pNode->PciBar[BarIndex].Properties.Physical = PciBar & ~0xF;
            pNode->PciBar[BarIndex].Properties.Size     = (~(Size & ~0xF)) + 1;
        }
    }

    // Return PCI BAR properties
    *pBarProperties = pNode->PciBar[BarIndex].Properties;

    // Display BAR properties if enabled
    if (pNode->PciBar[BarIndex].Properties.Size == 0)
    {
        DebugPrintf((
            "PCI BAR %d is disabled\n",
            BarIndex
            ));
    }
    else
    {
        DebugPrintf((
            "    PCI BAR %d: %08X\n",
            BarIndex, pNode->PciBar[BarIndex].Properties.BarValue
            ));

        DebugPrintf((
            "    Phys Addr: %08lX\n",
            (PLX_UINT_PTR)pNode->PciBar[BarIndex].Properties.Physical
            ));

        if (pNode->PciBar[BarIndex].Properties.Size >= (1 << 10))
        {
            DebugPrintf((
                "    Size     : %08lx  (%ld Kb)\n",
                (PLX_UINT_PTR)pNode->PciBar[BarIndex].Properties.Size,
                (PLX_UINT_PTR)pNode->PciBar[BarIndex].Properties.Size >> 10
                ));
        }
        else
        {
            DebugPrintf((
                "    Size     : %08lx  (%ld bytes)\n",
                (PLX_UINT_PTR)pNode->PciBar[BarIndex].Properties.Size,
                (PLX_UINT_PTR)pNode->PciBar[BarIndex].Properties.Size
                ));
        }

        DebugPrintf((
            "    Prefetch?: %s\n",
            (pNode->PciBar[BarIndex].Properties.bPrefetchable) ? "Yes" : "No"
            ));
    }

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxPciBarMap
 *
 * Description:  Map a PCI BAR Space into User virtual space
 *
 ******************************************************************************/
PLX_STATUS
PlxPciBarMap(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U8                BarIndex,
    VOID             *pUserVa
    )
{
    PLX_DEVICE_NODE  *pNode;
    PLX_USER_MAPPING *pObject;


    /*************************************************************
     * In the service driver, this function simply adds an entry
     * to a list with additional information needed when the
     * API calls mmap and Dispatch_mmap executes.
     ************************************************************/

    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    pObject =
        kmalloc(
            sizeof(PLX_USER_MAPPING),
            GFP_KERNEL
            );

    if (pObject == NULL)
    {
        DebugPrintf(("ERROR - memory allocation for map object failed\n"));
        return ApiInsufficientResources;
    }

    // Save the intended device node
    pObject->pNode = pNode;

    // Record the desired BAR index
    pObject->BarIndex = BarIndex;

    // Increment request counter
    pNode->MapRequestPending++;

    // Add to list of map objects
    spin_lock(
        &(pdx->Lock_MapParamsList)
        );

    list_add_tail(
        &(pObject->ListEntry),
        &(pdx->List_MapParams)
        );

    spin_unlock(
        &(pdx->Lock_MapParamsList)
        );

    DebugPrintf((
        "Added map object (%p) to list for node (%p)\n",
        pObject, pNode
        ));

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxPciBarUnmap
 *
 * Description:  Unmap a previously mapped PCI BAR Space from User virtual space
 *
 ******************************************************************************/
PLX_STATUS
PlxPciBarUnmap(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    VOID             *UserVa
    )
{
    // Handled at user API level in Linux
    return ApiUnsupportedFunction;
}




/*******************************************************************************
 *
 * Function   :  PlxEepromPresent
 *
 * Description:  Returns the state of the EEPROM as reported by the PLX device
 *
 ******************************************************************************/
PLX_STATUS
PlxEepromPresent(
    DEVICE_EXTENSION  *pdx,
    PLX_DEVICE_KEY    *pKey,
    PLX_EEPROM_STATUS *pStatus
    )
{
    U32              TempChip;
    PLX_DEVICE_NODE *pNode;


    // Default to no EEPROM present
    *pStatus = PLX_EEPROM_STATUS_NONE;

    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Generalize by device type
    if (((pNode->PlxChipType & 0xFF00) == 0x8500) ||
        ((pNode->PlxChipType & 0xFF00) == 0x8600))
    {
        TempChip = 0x8000;
    }
    else
    {
        TempChip = pNode->PlxChipType;
    }

    switch (TempChip)
    {
        case 0x6150:
        case 0x6152:
        case 0x6154:
        case 0x6156:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
            return Plx6000_EepromPresent(
                pNode,
                pStatus
                );

        case 0x8111:
        case 0x8112:
            return Plx8111_EepromPresent(
                pNode,
                pStatus
                );

        case 0x8114:
        case 0x8000:
            return Plx8000_EepromPresent(
                pNode,
                pStatus
                );
    }

    return ApiUnsupportedFunction;
}




/*******************************************************************************
 *
 * Function   :  PlxEepromProbe
 *
 * Description:  Probes for the presence of an EEPROM
 *
 ******************************************************************************/
PLX_STATUS
PlxEepromProbe(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    BOOLEAN          *pFlag
    )
{
    U16              OffsetProbe;
    U32              TempChip;
    U32              ValueRead;
    U32              ValueWrite;
    U32              ValueOriginal;
    PLX_STATUS       rc;
    PLX_DEVICE_NODE *pNode;


    // Default to no EEPROM present
    *pFlag = FALSE;

    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Generalize by device type
    if ((pNode->PlxChipType & 0xFF00) == 0x8600)
    {
        TempChip = 0x8600;
    }
    else
    {
        TempChip = pNode->PlxChipType;
    }

    // Determine EEPROM offset to use for probe (e.g. after CRC)
    switch (TempChip)
    {
        case 0x6150:
        case 0x6152:
        case 0x6154:
        case 0x6156:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
            OffsetProbe = 0x50;     // Use area outside of PLX portion
            break;

        case 0x8114:
            if (pNode->PlxRevision >= 0xBA)
                OffsetProbe = 0x3EC + sizeof(U32);
            else
                OffsetProbe = 0x378 + sizeof(U32);
            break;

        case 0x8508:
        case 0x8512:
        case 0x8517:
        case 0x8518:
            OffsetProbe = (0x78F * sizeof(U32)) + sizeof(U32);
            break;

        case 0x8516:
        case 0x8524:
        case 0x8532:
            OffsetProbe = (0xBE4 * sizeof(U32)) + sizeof(U32);
            break;

        case 0x8111:
        case 0x8112:
        case 0x8505:
        case 0x8509:
        case 0x8525:
        case 0x8533:
        case 0x8547:
        case 0x8548:
        case 0x8600:
            OffsetProbe = 0x10;     // No CRC, just use any reasonable address
            break;

        case 0:
        default:
            DebugPrintf((
                "ERROR - Not a supported PLX device (%04X)\n",
                pNode->PlxChipType
                ));
            return ApiUnsupportedFunction;
    }

    DebugPrintf(("Probing EEPROM at offset %02xh\n", OffsetProbe));

    // Get the current value
    rc =
        PlxEepromReadByOffset(
            pdx,
            pKey,
            OffsetProbe,
            &ValueOriginal
            );

    if (rc != ApiSuccess)
        return rc;

    // Prepare inverse value to write
    ValueWrite = ~(ValueOriginal);

    // Write inverse of original value
    rc =
        PlxEepromWriteByOffset(
            pdx,
            pKey,
            OffsetProbe,
            ValueWrite
            );

    if (rc != ApiSuccess)
        return rc;

    // Read updated value
    rc =
        PlxEepromReadByOffset(
            pdx,
            pKey,
            OffsetProbe,
            &ValueRead
            );

    if (rc != ApiSuccess)
        return rc;

    // Check if value was written properly
    if (ValueRead == ValueWrite)
    {
        DebugPrintf(("Probe detected an EEPROM present\n"));

        *pFlag = TRUE;

        // Restore the original value
        PlxEepromWriteByOffset(
            pdx,
            pKey,
            OffsetProbe,
            ValueOriginal
            );
    }
    else
    {
        DebugPrintf(("Probe did not detect an EEPROM\n"));
    }

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxEepromSetAddressWidth
 *
 * Description:  Sets a new EEPROM address width
 *
 ******************************************************************************/
PLX_STATUS
PlxEepromSetAddressWidth(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U8                width
    )
{
    U32              ChipType;
    PLX_STATUS       rc;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Verify the width
    switch (width)
    {
        case 1:
        case 2:
        case 3:
            break;

        default:
            DebugPrintf((
                "ERROR - Invalid address width (%d)\n",
                width
                ));
            return ApiInvalidData;
    }

    // Get chip type
    ChipType = pNode->PlxChipType;

    // Generalize by device type
    if (((ChipType & 0xFF00) == 0x8500) ||
        ((ChipType & 0xFF00) == 0x8600))
    {
        ChipType = 0x8000;
    }

    switch (ChipType)
    {
        case 0x8111:
        case 0x8112:
            rc = ApiSuccess;
            break;

        case 0x8000:
            rc =
                Plx8000_EepromSetAddressWidth(
                    pNode,
                    width
                    );
            break;

        default:
            DebugPrintf((
                "ERROR - Chip (%04X) does not support address width override\n",
                pNode->PlxChipType
                ));
            return ApiUnsupportedFunction;
    }

    if (rc == ApiSuccess)
    {
        // Update the width
        pNode->Default_EepWidth = width;

        DebugPrintf((
            "Set EEPROM address width to %d bytes\n",
            width
            ));
    }

    return rc;
}




/*******************************************************************************
 *
 * Function   :  PlxEepromCrcGet
 *
 * Description:  Returns the EEPROM CRC and its status
 *
 ******************************************************************************/
PLX_STATUS
PlxEepromCrcGet(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U32              *pCrc,
    U8               *pCrcStatus
    )
{
    PLX_DEVICE_NODE *pNode;


    // Clear return value
    *pCrc = 0;

    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    switch (pNode->PlxChipType)
    {
        case 0x8114:
        case 0x8508:
        case 0x8512:
        case 0x8516:
        case 0x8517:
        case 0x8518:
        case 0x8524:
        case 0x8532:
            return Plx8000_EepromCrcGet(
                pNode,
                pCrc,
                pCrcStatus
                );

        default:
            // Devices don't support CRC
            break;
    }

    return ApiUnsupportedFunction;
}




/*******************************************************************************
 *
 * Function   :  PlxEepromCrcUpdate
 *
 * Description:  Updates the EEPROM CRC
 *
 ******************************************************************************/
PLX_STATUS
PlxEepromCrcUpdate(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U32              *pCrc,
    BOOLEAN           bUpdateEeprom
    )
{
    PLX_DEVICE_NODE *pNode;


    // Clear return value
    *pCrc = 0;

    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    switch (pNode->PlxChipType)
    {
        case 0x8114:
        case 0x8508:
        case 0x8512:
        case 0x8516:
        case 0x8517:
        case 0x8518:
        case 0x8524:
        case 0x8532:
            return Plx8000_EepromCrcUpdate(
                pNode,
                pCrc,
                bUpdateEeprom
                );

        default:
            // Devices don't support CRC
            break;
    }

    return ApiUnsupportedFunction;
}




/*******************************************************************************
 *
 * Function   :  PlxEepromReadByOffset
 *
 * Description:  Read a 32-bit value from the EEPROM at a specified offset
 *
 ******************************************************************************/
PLX_STATUS
PlxEepromReadByOffset(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U16               offset,
    U32              *pValue
    )
{
    U32              TempChip;
    PLX_DEVICE_NODE *pNode;


    // Set default return value
    *pValue = 0;

    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Make sure offset is aligned on 32-bit boundary
    if (offset & (3 << 0))
    {
        return ApiInvalidOffset;
    }

    // Generalize by device type
    if (((pNode->PlxChipType & 0xFF00) == 0x8500) ||
        ((pNode->PlxChipType & 0xFF00) == 0x8600))
    {
        TempChip = 0x8000;
    }
    else
    {
        TempChip = pNode->PlxChipType;
    }

    switch (TempChip)
    {
        case 0x6150:
        case 0x6152:
        case 0x6154:
        case 0x6156:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
        case 0x8111:
        case 0x8112:
            /******************************************
             * For devices that do not support 32-bit
             * EEPROM accesses, use 2 16-bit accesses
             *****************************************/
            PlxEepromReadByOffset_16(
                pdx,
                pKey,
                offset,
                (U16*)pValue
                );

            PlxEepromReadByOffset_16(
                pdx,
                pKey,
                offset + sizeof(U16),
                (U16*)((U8*)pValue + sizeof(U16))
                );
            break;

        case 0x8114:
        case 0x8000:
            return Plx8000_EepromReadByOffset(
                pNode,
                offset,
                pValue
                );

        case 0x6140:  // 6140 does not support an EEPROM
        default:
            return ApiUnsupportedFunction;
    }

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxEepromWriteByOffset
 *
 * Description:  Write a 32-bit value to the EEPROM at a specified offset
 *
 ******************************************************************************/
PLX_STATUS
PlxEepromWriteByOffset(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U16               offset,
    U32               value
    )
{
    U32              TempChip;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Make sure offset is aligned on 32-bit boundary
    if (offset & (3 << 0))
    {
        return ApiInvalidOffset;
    }

    // Generalize by device type
    if (((pNode->PlxChipType & 0xFF00) == 0x8500) ||
        ((pNode->PlxChipType & 0xFF00) == 0x8600))
    {
        TempChip = 0x8000;
    }
    else
    {
        TempChip = pNode->PlxChipType;
    }

    switch (TempChip)
    {
        case 0x6150:
        case 0x6152:
        case 0x6154:
        case 0x6156:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
        case 0x8111:
        case 0x8112:
            /******************************************
             * For devices that do not support 32-bit
             * EEPROM accesses, use 2 16-bit accesses
             *****************************************/
            PlxEepromWriteByOffset_16(
                pdx,
                pKey,
                offset,
                (U16)value
                );

            PlxEepromWriteByOffset_16(
                pdx,
                pKey,
                offset + sizeof(U16),
                (U16)(value >> 16)
                );
            break;

        case 0x8114:
        case 0x8000:
            return Plx8000_EepromWriteByOffset(
                pNode,
                offset,
                value
                );

        case 0x6140:  // 6140 does not support an EEPROM
        default:
            return ApiUnsupportedFunction;
    }

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxEepromReadByOffset_16
 *
 * Description:  Read a 16-bit value from the EEPROM at a specified offset
 *
 ******************************************************************************/
PLX_STATUS
PlxEepromReadByOffset_16(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U16               offset,
    U16              *pValue
    )
{
    U32              Value_32;
    U32              TempChip;
    PLX_STATUS       rc;
    PLX_DEVICE_NODE *pNode;


    // Set default return value
    *pValue = 0;

    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Make sure offset is aligned on 16-bit boundary
    if (offset & (1 << 0))
    {
        return ApiInvalidOffset;
    }

    // Generalize by device type
    if (((pNode->PlxChipType & 0xFF00) == 0x8500) ||
        ((pNode->PlxChipType & 0xFF00) == 0x8600))
    {
        TempChip = 0x8000;
    }
    else
    {
        TempChip = pNode->PlxChipType;
    }

    switch (TempChip)
    {
        case 0x6150:
        case 0x6152:
        case 0x6154:
        case 0x6156:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
            return Plx6000_EepromReadByOffset_16(
                pNode,
                offset,
                pValue
                );

        case 0x8111:
        case 0x8112:
            return Plx8111_EepromReadByOffset_16(
                pNode,
                offset,
                pValue
                );

        case 0x8114:
        case 0x8000:
            /******************************************
             * For devices that do not support 16-bit
             * EEPROM accesses, use 32-bit access
             *****************************************/

            // Get current 32-bit value
            rc =
                PlxEepromReadByOffset(
                    pdx,
                    pKey,
                    (offset & ~0x3),
                    &Value_32
                    );

            if (rc != ApiSuccess)
                return rc;

            // Return desired 16-bit portion
            if ((offset & 0x3) == 2)
            {
                *pValue = (U16)(Value_32 >> 16);
            }
            else
            {
                *pValue = (U16)(Value_32);
            }
            break;

        case 0x6140:  // 6140 does not support an EEPROM
        default:
            DebugPrintf((
                "ERROR - Device (%04X) does not support 16-bit EEPROM access\n",
                pNode->PlxChipType
                ));
            return ApiUnsupportedFunction;
    }

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxEepromWriteByOffset_16
 *
 * Description:  Write a 16-bit value to the EEPROM at a specified offset
 *
 ******************************************************************************/
PLX_STATUS
PlxEepromWriteByOffset_16(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    U16               offset,
    U16               value
    )
{
    U32              Value_32;
    U32              TempChip;
    PLX_STATUS       rc;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Make sure offset is aligned on 16-bit boundary
    if (offset & (1 << 0))
    {
        return ApiInvalidOffset;
    }

    // Generalize by device type
    if (((pNode->PlxChipType & 0xFF00) == 0x8500) ||
        ((pNode->PlxChipType & 0xFF00) == 0x8600))
    {
        TempChip = 0x8000;
    }
    else
    {
        TempChip = pNode->PlxChipType;
    }

    switch (TempChip)
    {
        case 0x6150:
        case 0x6152:
        case 0x6154:
        case 0x6156:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
            return Plx6000_EepromWriteByOffset_16(
                pNode,
                offset,
                value
                );

        case 0x8111:
        case 0x8112:
            return Plx8111_EepromWriteByOffset_16(
                pNode,
                offset,
                value
                );

        case 0x8114:
        case 0x8000:
            /******************************************
             * For devices that do not support 16-bit
             * EEPROM accesses, use 32-bit access
             *****************************************/

            // Get current 32-bit value
            rc =
                PlxEepromReadByOffset(
                    pdx,
                    pKey,
                    (offset & ~0x3),
                    &Value_32
                    );

            if (rc != ApiSuccess)
                return rc;

            // Insert new 16-bit value in correct location
            if ((offset & 0x3) == 2)
            {
                Value_32 = ((U32)value << 16) | (Value_32 & 0xFFFF);
            }
            else
            {
                Value_32 = ((U32)value) | (Value_32 & 0xFFFF0000);
            }

            // Update EEPROM
            return PlxEepromWriteByOffset(
                pdx,
                pKey,
                (offset & ~0x3),
                Value_32
                );

        case 0x6140:  // 6140 does not support an EEPROM
        default:
            DebugPrintf((
                "ERROR - Device (%04X) does not support 16-bit EEPROM access\n",
                pNode->PlxChipType
                ));
            return ApiUnsupportedFunction;
    }

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxPciIoPortTransfer
 *
 * Description:  Read or Write from/to an I/O port
 *
 ******************************************************************************/
PLX_STATUS
PlxPciIoPortTransfer(
    U64              IoPort,
    VOID            *pBuffer,
    U32              SizeInBytes,
    PLX_ACCESS_TYPE  AccessType,
    BOOLEAN          bReadOperation
    )
{
    U8 AccessSize;


    if (pBuffer == NULL)
        return ApiNullParam;

    // Verify size & type
    switch (AccessType)
    {
        case BitSize8:
            AccessSize = sizeof(U8);
            break;

        case BitSize16:
            if (IoPort & (1 << 0))
            {
                DebugPrintf(("ERROR - I/O port not aligned on 16-bit boundary\n"));
                return ApiInvalidAddress;
            }

            if (SizeInBytes & (1 << 0))
            {
                DebugPrintf(("ERROR - Byte count not aligned on 16-bit boundary\n"));
                return ApiInvalidSize;
            }
            AccessSize = sizeof(U16);
            break;

        case BitSize32:
            if (IoPort & 0x3)
            {
                DebugPrintf(("ERROR - I/O port not aligned on 32-bit boundary\n"));
                return ApiInvalidAddress;
            }

            if (SizeInBytes & 0x3)
            {
                DebugPrintf(("ERROR - Byte count not aligned on 32-bit boundary\n"));
                return ApiInvalidSize;
            }
            AccessSize = sizeof(U32);
            break;

        default:
            return ApiInvalidAccessType;
    }

    while (SizeInBytes)
    {
        if (bReadOperation)
        {
            switch (AccessType)
            {
                case BitSize8:
                    *(U8*)pBuffer = IO_PORT_READ_8( IoPort );
                    break;

                case BitSize16:
                    *(U16*)pBuffer = IO_PORT_READ_16( IoPort );
                    break;

                case BitSize32:
                    *(U32*)pBuffer = IO_PORT_READ_32( IoPort );
                    break;

                default:
                    // Added to avoid compiler warnings
                    break;
            }
        }
        else
        {
            switch (AccessType)
            {
                case BitSize8:
                    IO_PORT_WRITE_8(
                        IoPort,
                        *(U8*)pBuffer
                        );
                    break;

                case BitSize16:
                    IO_PORT_WRITE_16(
                        IoPort,
                        *(U16*)pBuffer
                        );
                    break;

                case BitSize32:
                    IO_PORT_WRITE_32(
                        IoPort,
                        *(U32*)pBuffer
                        );
                    break;

                default:
                    // Added to avoid compiler warnings
                    break;
            }
        }

        // Adjust pointer & byte count
        pBuffer      = (VOID*)((PLX_UINT_PTR)pBuffer + AccessSize);
        SizeInBytes -= AccessSize;
    }

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxPciPerformanceInitializeProperties
 *
 * Description:  Initializes the performance properties for a device
 *
 ******************************************************************************/
PLX_STATUS
PlxPciPerformanceInitializeProperties(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    PLX_PERF_PROP    *pPerfProp
    )
{
    PLX_PORT_PROP    PortProp;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Verify supported device
    if (((pNode->PlxChipType & 0xFF00) != 0x8600) &&
         (pNode->PlxChipType != 0x8511) &&
         (pNode->PlxChipType != 0x8513) &&
         (pNode->PlxChipType != 0x8519))
    {
        DebugPrintf(("ERROR - Device (%04X) doesn't support performance counters\n", pNode->PlxChipType));
        return ApiUnsupportedFunction;
    }

    // Clear performance object
    RtlZeroMemory( pPerfProp, sizeof(PLX_PERF_PROP) );

    // Get port properties
    PlxGetPortProperties(
        pdx,
        pKey,
        &PortProp
        );

    if (PortProp.PortNumber >= 16)
    {
        DebugPrintf(("ERROR - Port number exceeds maximum (%d)\n", (16-1)));
        return ApiUnsupportedFunction;
    }

    // Store relevant port properties for later calculations
    pPerfProp->PortNumber = PortProp.PortNumber;
    pPerfProp->LinkWidth  = PortProp.LinkWidth;
    pPerfProp->LinkSpeed  = PortProp.LinkSpeed;

    // Determine station and port number within station
    pPerfProp->Station     = (U8)(PortProp.PortNumber / PERF_PORTS_PER_STATION);
    pPerfProp->StationPort = (U8)(PortProp.PortNumber - (pPerfProp->Station * PERF_PORTS_PER_STATION));

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxPciPerformanceMonitorControl
 *
 * Description:  Controls PLX Performance Monitor
 *
 ******************************************************************************/
PLX_STATUS
PlxPciPerformanceMonitorControl(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    PLX_PERF_CMD      command
    )
{
    U32              RegValue;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Verify supported device
    if (((pNode->PlxChipType & 0xFF00) != 0x8600) &&
         (pNode->PlxChipType != 0x8511) &&
         (pNode->PlxChipType != 0x8513) &&
         (pNode->PlxChipType != 0x8519))
    {
        DebugPrintf(("ERROR - Device (%04X) doesn't support performance counters\n", pNode->PlxChipType));
        return ApiUnsupportedFunction;
    }

    switch (command)
    {
        case PLX_PERF_CMD_START:
            DebugPrintf(("Reset & enable monitor with infinite sampling\n"));
            RegValue = (1 << 31) | (1 << 30) | (1 << 28) | (1 << 27);
            break;

        case PLX_PERF_CMD_STOP:
            DebugPrintf(("Reset & disable monitor\n"));
            RegValue = (1 << 30);
            break;

        default:
            return ApiInvalidData;
    }

    // Update monitor
    PLX_8000_REG_WRITE(
        pNode,
        0x568,
        RegValue
        );

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxPciPerformanceResetCounters
 *
 * Description:  Resets the internal performance counters
 *
 ******************************************************************************/
PLX_STATUS
PlxPciPerformanceResetCounters(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey
    )
{
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Verify supported device
    if (((pNode->PlxChipType & 0xFF00) != 0x8600) &&
         (pNode->PlxChipType != 0x8511) &&
         (pNode->PlxChipType != 0x8513) &&
         (pNode->PlxChipType != 0x8519))
    {
        DebugPrintf(("ERROR - Device (%04X) doesn't support performance counters\n", pNode->PlxChipType));
        return ApiUnsupportedFunction;
    }

    // Reset (bit 30) & enable monitor (bit 31) & infinite sampling (bit 28)
    PLX_8000_REG_WRITE(
        pNode,
        0x568,
        (1 << 31) | (1 << 30) | (1 << 28) | (1 << 27)
        );

    return ApiSuccess;
}




/*******************************************************************************
 *
 * Function   :  PlxPciPerformanceGetCounters
 *
 * Description:  Retrieves a snapshot of the Performance Counters for selected ports
 *
 ******************************************************************************/
PLX_STATUS
PlxPciPerformanceGetCounters(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey,
    PLX_PERF_PROP    *pPerfProps,
    U8                NumOfObjects
    )
{
    U8               i;
    U8               NumPorts;
    U8               NumCounters;
    U16              index;
    U16              IndexBase;
    U32              RegValue;
    U32              Counters[16 * 14];     // Max is 16 ports x 14 counters/port
    BOOLEAN          bLegacyCounters;
    PLX_DEVICE_NODE *pNode;


    pNode =
        GetDeviceNodeFromKey(
            pdx,
            pKey
            );

    if (pNode == NULL)
    {
        DebugPrintf(("ERROR - Invalid Device node\n"));
        return ApiInvalidDeviceInfo;
    }

    // Verify supported device
    if (((pNode->PlxChipType & 0xFF00) != 0x8600) &&
         (pNode->PlxChipType != 0x8511) &&
         (pNode->PlxChipType != 0x8513) &&
         (pNode->PlxChipType != 0x8519))
    {
        DebugPrintf(("ERROR - Device (%04X) doesn't support performance counters\n", pNode->PlxChipType));
        return ApiUnsupportedFunction;
    }

    // Setup parameters for reading counters
    switch (pNode->PlxChipType)
    {
        case 0x8612:
        case 0x8616:
        case 0x8624:
        case 0x8632:
        case 0x8647:
        case 0x8648:
            NumPorts        = 12;
            NumCounters     = 14;
            bLegacyCounters = TRUE;
            break;

        default:
            NumPorts        = 16;
            NumCounters     = 13;
            bLegacyCounters = FALSE;
            break;
    }

    // RAM control default - Set capture type [5:4] to 2 & enable RAM (bit 0)
    RegValue = (2 << 4) | (1 << 0);

    // Reset RAM read pointer (bit 2)
    PLX_8000_REG_WRITE(
        pNode,
        0x618,
        RegValue | (1 << 2)
        );

    // Read in all counters
    i = 0;
    while (i < (NumPorts * NumCounters))
    {
        // Get next counter
        Counters[i] =
            PLX_8000_REG_READ(
                pNode,
                0x628
                );

        // Jump to next counter
        i++;
    }

    // Assign counter values to enabled ports
    i = 0;
    while (i < NumOfObjects)
    {
        // Save current values to previous
        RtlCopyMemory(
            &(pPerfProps[i].Prev_IngressPostedHeader),
            &(pPerfProps[i].IngressPostedHeader),
            NumCounters * sizeof(U32)
            );

        // Calculate starting index for counters
        if (bLegacyCounters)
        {
            // Legacy counters are based by station number
            IndexBase = pPerfProps[i].Station * (NumCounters * PERF_PORTS_PER_STATION);
        }
        else
        {
            IndexBase = 0;
        }

        // Get Ingress counters (5 DW/port)
        if (bLegacyCounters)
        {
            // Legacy ingress counters start at index 0 from base
            index = IndexBase + 0 + (pPerfProps[i].StationPort * 5);
        }
        else
        {
            // Ingress counters are in port order starting at offset 0
            index = IndexBase + 0 + (pPerfProps[i].PortNumber * 5);
        }
        pPerfProps[i].IngressPostedHeader = Counters[index + 0];
        pPerfProps[i].IngressPostedDW     = Counters[index + 1];
        pPerfProps[i].IngressNonpostedDW  = Counters[index + 2];
        pPerfProps[i].IngressCplHeader    = Counters[index + 3];
        pPerfProps[i].IngressCplDW        = Counters[index + 4];

        // Get Egress counters (5 DW/port)
        if (bLegacyCounters)
        {
            // Legacy engress counters start after ingress (index 20) from base
            index = IndexBase + 20 + (pPerfProps[i].StationPort * 5);
        }
        else
        {
            // Engress counters start after ingress (index 80) from base
            index = IndexBase + 80 + (pPerfProps[i].PortNumber * 5);
        }
        pPerfProps[i].EgressPostedHeader = Counters[index + 0];
        pPerfProps[i].EgressPostedDW     = Counters[index + 1];
        pPerfProps[i].EgressNonpostedDW  = Counters[index + 2];
        pPerfProps[i].EgressCplHeader    = Counters[index + 3];
        pPerfProps[i].EgressCplDW        = Counters[index + 4];

        // Get DLLP Ingress counters (1 DW/port)
        if (bLegacyCounters)
        {
            // Legacy DLLP Ingress counters start after egress (index 40)
            index = IndexBase + 40 + (pPerfProps[i].StationPort * 1);
        }
        else
        {
            // DLLP ingress counters start after egress (index 160)
            index = IndexBase + 160 + (pPerfProps[i].PortNumber / 2);

            // Odd port counters follow even DLLP egress counters (index 176)
            if (pPerfProps[i].PortNumber & (1 << 0))
                index += NumPorts;
        }
        pPerfProps[i].IngressDllp = Counters[index];

        // Get DLLP Egress counters (1 DW/port)
        if (bLegacyCounters)
        {
            // Legacy DLLP Egress counters start after DLLP ingress (index 44)
            index = IndexBase + 44 + (pPerfProps[i].StationPort * 1);
        }
        else
        {
            // DLLP egress counters start after even DLLP ingress (index 168)
            index = IndexBase + 168 + (pPerfProps[i].PortNumber / 2);

            // Odd port counters follow odd DLLP ingress counters (index 184)
            if (pPerfProps[i].PortNumber & (1 << 0))
                index += NumPorts;
        }
        pPerfProps[i].EgressDllp = Counters[index];

        // Get PHY counters (1 DW/port)
        if (bLegacyCounters)
        {
            // Legacy PHY ingress & egress counters are at indexes 48 & 52
            index = IndexBase + 48 + (pPerfProps[i].StationPort * 1);
            pPerfProps[i].IngressPhy = Counters[index];

            index = IndexBase + 52 + (pPerfProps[i].StationPort * 1);
            pPerfProps[i].EgressPhy = Counters[index];
        }
        else
        {
            // PHY counter starts after DLLP egress counters (index 192)
            index = IndexBase + 192 + pPerfProps[i].PortNumber;

            // Only one PHY counter exists, so use same value for Ingress & Egress
            pPerfProps[i].IngressPhy = Counters[index];
            pPerfProps[i].EgressPhy  = pPerfProps[i].IngressPhy;
        }

        // Go to next performance object
        i++;
    }

    return ApiSuccess;
}
