/*******************************************************************************
 * Copyright (c) PLX Technology, Inc.
 *
 * PLX Technology Inc. licenses this source file under the GNU Lesser General Public
 * License (LGPL) version 2.  This source file may be modified or redistributed
 * under the terms of the LGPL and without express permission from PLX Technology.
 *
 * PLX Technology, Inc. provides this software AS IS, WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING, WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  PLX makes no guarantee
 * or representations regarding the use of, or the results of the use of,
 * the software and documentation in terms of correctness, accuracy,
 * reliability, currentness, or otherwise; and you rely on the software,
 * documentation and results solely at your own risk.
 *
 * IN NO EVENT SHALL PLX BE LIABLE FOR ANY LOSS OF USE, LOSS OF BUSINESS,
 * LOSS OF PROFITS, INDIRECT, INCIDENTAL, SPECIAL OR CONSEQUENTIAL DAMAGES
 * OF ANY KIND.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * File Name:
 *
 *      Eep_6000.c
 *
 * Description:
 *
 *      This file contains 6000-series EEPROM support functions
 *
 * Revision History:
 *
 *      12-01-07 : PLX SDK v5.20
 *
 ******************************************************************************/


#include "Eep_6000.h"
#include "PciFunc.h"
#include "SuppFunc.h"




/******************************************************************************
 *
 * Function   :  Plx6000_EepromPresent
 *
 * Description:  Determine if an EEPROM is present on the device
 *
 ******************************************************************************/
PLX_STATUS
Plx6000_EepromPresent(
    PLX_DEVICE_NODE   *pNode,
    PLX_EEPROM_STATUS *pStatus
    )
{
    U16 Value_Test;
    U16 Value_Original;


    /**************************************************************
     * Since the PCI 6000 series do not report whether an
     * EEPROM is present or not, the driver checks for one
     * as follows:
     *
     *  - Read an EEPROM location outside PLX chip reserved area
     *  - Write the inverse of the value to the location
     *  - Read back and compare
     *  - If value was modified, an EEPROM is present
     *  - Restore original value
     *************************************************************/

    // Get original value from unused location
    Plx6000_EepromReadByOffset_16(
        pNode,
        EEPROM_6000_TEST_OFFSET,
        &Value_Original
        );

    // Write inverse of original value
    Plx6000_EepromWriteByOffset_16(
        pNode,
        EEPROM_6000_TEST_OFFSET,
        (U16)~Value_Original
        );

    // Read updated value
    Plx6000_EepromReadByOffset_16(
        pNode,
        EEPROM_6000_TEST_OFFSET,
        &Value_Test
        );

    // Restore original value
    Plx6000_EepromWriteByOffset_16(
        pNode,
        EEPROM_6000_TEST_OFFSET,
        Value_Original
        );

    // Check if value was written properly
    if (Value_Test == (U16)~Value_Original)
    {
        *pStatus = PLX_EEPROM_STATUS_VALID;
    }
    else
    {
        *pStatus = PLX_EEPROM_STATUS_NONE;
    }

    return ApiSuccess;
}




/******************************************************************************
 *
 * Function   :  Plx6000_EepromReadByOffset_16
 *
 * Description:  Read a 16-bit value from the EEPROM at a specified offset
 *
 ******************************************************************************/
PLX_STATUS
Plx6000_EepromReadByOffset_16(
    PLX_DEVICE_NODE *pNode,
    U16              offset,
    U16             *pValue
    )
{
    U16 Offset_EepromCtrl;
    U32 RegSave;
    U32 RegValue;


    switch (pNode->PlxChipType)
    {
        case 0x6152:
        case 0x6156:
            Offset_EepromCtrl = 0xC8;
            break;

        case 0x6150:
        case 0x6154:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
            Offset_EepromCtrl = 0x54;
            break;

        case 0x6140:    // No EEPROM support
        default:
            DebugPrintf((
                "ERROR - Unsupported PLX chip type (%04X)\n",
                pNode->PlxChipType
                ));
            return ApiUnsupportedFunction;
    }

    // For Non-transparent mode chips, enable shadow registers
    if ((pNode->PlxChipType == 0x6254) ||
        (pNode->PlxChipType == 0x6540))
    {
        // Save the chip control register
        PLX_PCI_REG_READ(
            pNode,
            0xD8,
            &RegSave
            );

        // Enable shadow register access
        PLX_PCI_REG_WRITE(
            pNode,
            0xD8,
            RegSave | (1 << 6)
            );
    }

    // Offset can only be 8-bits
    offset = offset & 0xFF;

    // Prepare EEPROM read command
    RegValue = ((U32)offset << 8) | (1 << 0);

    // Write EEPROM command
    PLX_PCI_REG_WRITE(
        pNode,
        Offset_EepromCtrl,
        RegValue
        );

    // Insert a small delay
    Plx_sleep(50);

    // Read data
    PLX_PCI_REG_READ(
        pNode,
        Offset_EepromCtrl,
        &RegValue
        );

    // Store data
    *pValue = (U16)(RegValue >> 16);

    if ((pNode->PlxChipType == 0x6254) ||
        (pNode->PlxChipType == 0x6540))
    {
        // Restore chip control register
        PLX_PCI_REG_WRITE(
            pNode,
            0xD8,
            RegSave
            );
    }

    return ApiSuccess;
}




/******************************************************************************
 *
 * Function   :  Plx6000_EepromWriteByOffset_16
 *
 * Description:  Write a 16-bit value to the EEPROM at a specified offset
 *
 ******************************************************************************/
PLX_STATUS
Plx6000_EepromWriteByOffset_16(
    PLX_DEVICE_NODE *pNode,
    U16              offset,
    U16              value
    )
{
    U16 Offset_EepromCtrl;
    U32 RegSave;
    U32 RegValue;


    switch (pNode->PlxChipType)
    {
        case 0x6152:
        case 0x6156:
            Offset_EepromCtrl = 0xC8;
            break;

        case 0x6150:
        case 0x6154:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
            Offset_EepromCtrl = 0x54;
            break;

        case 0x6140:    // No EEPROM support
        default:
            DebugPrintf((
                "ERROR - Unsupported PLX chip type (%04X)\n",
                pNode->PlxChipType
                ));
            return ApiUnsupportedFunction;
    }

    // For Non-transparent mode chips, enable shadow registers
    if ((pNode->PlxChipType == 0x6254) ||
        (pNode->PlxChipType == 0x6540))
    {
        // Save the chip control register
        PLX_PCI_REG_READ(
            pNode,
            0xD8,
            &RegSave
            );

        // Enable shadow register access
        PLX_PCI_REG_WRITE(
            pNode,
            0xD8,
            RegSave | (1 << 6)
            );
    }

    // Offset can only be 8-bits
    offset = offset & 0xFF;

    // Prepare EEPROM write command & data
    RegValue = (value << 16) | ((U32)offset << 8) | (1 << 1) | (1 << 0);

    // Write EEPROM command
    PLX_PCI_REG_WRITE(
        pNode,
        Offset_EepromCtrl,
        RegValue
        );

    // Insert a small delay
    Plx_sleep(100);

    if ((pNode->PlxChipType == 0x6254) ||
        (pNode->PlxChipType == 0x6540))
    {
        // Restore chip control register
        PLX_PCI_REG_WRITE(
            pNode,
            0xD8,
            RegSave
            );
    }

    return ApiSuccess;
}
