/*******************************************************************************
 * Copyright (c) PLX Technology, Inc.
 *
 * PLX Technology Inc. licenses this source file under the GNU Lesser General Public
 * License (LGPL) version 2.  This source file may be modified or redistributed
 * under the terms of the LGPL and without express permission from PLX Technology.
 *
 * PLX Technology, Inc. provides this software AS IS, WITHOUT ANY WARRANTY,
 * EXPRESS OR IMPLIED, INCLUDING, WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  PLX makes no guarantee
 * or representations regarding the use of, or the results of the use of,
 * the software and documentation in terms of correctness, accuracy,
 * reliability, currentness, or otherwise; and you rely on the software,
 * documentation and results solely at your own risk.
 *
 * IN NO EVENT SHALL PLX BE LIABLE FOR ANY LOSS OF USE, LOSS OF BUSINESS,
 * LOSS OF PROFITS, INDIRECT, INCIDENTAL, SPECIAL OR CONSEQUENTIAL DAMAGES
 * OF ANY KIND.
 *
 ******************************************************************************/

/******************************************************************************
 *
 * File Name:
 *
 *      SuppFunc.c
 *
 * Description:
 *
 *      Additional support functions
 *
 * Revision History:
 *
 *      12-01-07 : PLX SDK v5.20
 *
 ******************************************************************************/


#include <linux/delay.h>
#include "ApiFunc.h"
#include "PciFunc.h"
#include "SuppFunc.h"




/*******************************************************************************
 *
 * Function   :  Plx_sleep
 *
 * Description:  Function as a normal sleep. Parameter is in millisecond
 *
 ******************************************************************************/
VOID
Plx_sleep(
    U32 delay
    )
{
    mdelay(
        delay
        );
}




/*******************************************************************************
 *
 * Function   :  Plx_pow_int
 *
 * Description:  A PLX support routine to compute x^y (integers)
 *
 ******************************************************************************/
U32
Plx_pow_int(
    U32 x,
    U32 y
    )
{
    U32 value;


    // Check for '0' exponent
    if (y == 0)
        return 1;

    // Check for base of '0'
    if (x == 0)
        return 0;

    // Set start value
    value = 1;

    // Perform calculation
    while (y--)
    {
        value = value * x;
    }

    return value;
}




/*******************************************************************************
 *
 * Function   :  PlxGetExtendedCapabilityOffset
 *
 * Description:  Scans the capability list to search for a specific capability
 *
 ******************************************************************************/
U16
PlxGetExtendedCapabilityOffset(
    PLX_DEVICE_NODE *pNode,
    U16              CapabilityId
    )
{
    U16 Offset_Cap;
    U32 RegValue;


    // Verify PCI header is not Type 2
    if (pNode->PciHeaderType == 2)
    {
        DebugPrintf(("Device header is PCI Type 2 (Cardbus) - PCI extended capabilities not supported\n"));
        return 0;
    }

    // Get offset of first capability
    PLX_PCI_REG_READ(
        pNode,
        0x34,           // PCI capabilities pointer
        &RegValue
        );

    // If link is down, PCI reg accesses will fail
    if (RegValue == (U32)-1)
        return 0;

    // Set first capability
    Offset_Cap = (U16)RegValue;

    // Traverse capability list searching for desired ID
    while ((Offset_Cap != 0) && (RegValue != (U32)-1))
    {
        // Get next capability
        PLX_PCI_REG_READ(
            pNode,
            Offset_Cap,
            &RegValue
            );

        if ((U8)RegValue == (U8)CapabilityId)
        {
            // Capability found, return base offset
            return Offset_Cap;
        }

        // Jump to next capability
        Offset_Cap = (U16)((RegValue >> 8) & 0xFF);
    }

    // Capability not found
    return 0;
}




/*******************************************************************************
 *
 * Function   :  PlxPciBarResourceMap
 *
 * Description:  Maps a PCI BAR resource into kernel space
 *
 ******************************************************************************/
int
PlxPciBarResourceMap(
    PLX_DEVICE_NODE *pNode,
    U8               BarIndex
    )
{
    PLX_PCI_BAR_PROP BarProp;


    // Verify BAR index
    switch (pNode->PciHeaderType)
    {
        case 0:
            if ((BarIndex != 0) && (BarIndex > 5))
            {
                DebugPrintf(("Invalid PCI BAR (%d) specified\n", BarIndex));
                return (-1);
            }
            break;

        case 1:
            if ((BarIndex != 0) && (BarIndex != 1))
            {
                DebugPrintf(("PCI BAR %d does not exist on PCI type 1 headers\n", BarIndex));
                return (-1);
            }
            break;

        default:
            DebugPrintf(("PCI Header Type %d is not supported for PCI BAR mappings\n", pNode->PciHeaderType));
            return (-1);
    }

    // Check if BAR already mapped
    if (pNode->PciBar[BarIndex].pVa != NULL)
    {
        return 0;
    }

    DebugPrintf((
        "Mapping PCI BAR %d to kernel space...\n",
        BarIndex
        ));

    // Get BAR properties
    if (PlxPciBarProperties(
            pNode->pdx,
            &pNode->Key,
            BarIndex,
            &BarProp
            ) != ApiSuccess)
    {
        return (-1);
    }

    // Verify BAR is valid and is memory space
    if ((BarProp.Physical == 0) || (BarProp.bIoSpace))
    {
        return (-1);
    }

    // Get a kernel-mapped virtual address
    pNode->PciBar[BarIndex].pVa =
        ioremap(
            pNode->PciBar[BarIndex].Properties.Physical,
            pNode->PciBar[BarIndex].Properties.Size
            );

    if (pNode->PciBar[BarIndex].pVa == NULL)
    {
        ErrorPrintf((
            "    Kernel VA: ERROR - Unable to map space to Kernel VA\n"
            ));
        return (-1);
    }

    DebugPrintf((
        "    Kernel VA: %p\n",
        pNode->PciBar[BarIndex].pVa
        ));

    // Mark that node has BARs mapped
    pNode->bBarsMapped = TRUE;

    return 0;
}




/*******************************************************************************
 *
 * Function   :  PlxPciBarResourcesUnmap
 *
 * Description:  Unmap mapped PCI BAR memory for a device
 *
 ******************************************************************************/
int
PlxPciBarResourcesUnmap(
    PLX_DEVICE_NODE *pNode,
    U8               BarIndex
    )
{
    U8 i;
    U8 MaxBar;


    if (pNode->PciHeaderType == 0)
        MaxBar = PCI_NUM_BARS_TYPE_00;
    else if (pNode->PciHeaderType == 1)
        MaxBar = PCI_NUM_BARS_TYPE_01;
    else
        return (-1);

    for (i = 0; i < MaxBar; i++)
    {
        if ((BarIndex == (U8)-1) || (BarIndex == i))
        {
            // Unmap the space from Kernel space if previously mapped
            if (pNode->PciBar[i].pVa == NULL)
            {
                if (BarIndex != (U8)-1)
                    return (-1);
            }
            else
            {
                DebugPrintf((
                    "Unmapping BAR %d (VA=%p)...\n",
                    i, pNode->PciBar[i].pVa
                    ));

                // Unmap from kernel space
                iounmap(
                    pNode->PciBar[i].pVa
                    );

                // Clear BAR information
                RtlZeroMemory(
                    &(pNode->PciBar[i]),
                    sizeof(PLX_PCI_BAR_INFO)
                    );
            }
        }
    }

    return 0;
}




/*******************************************************************************
 *
 * Function   :  PlxResourcesReleaseAll
 *
 * Description:  Releases resources for any device that has aquired any
 *
 ******************************************************************************/
int
PlxResourcesReleaseAll(
    DEVICE_EXTENSION *pdx
    )
{
    struct list_head *pEntry;
    PLX_DEVICE_NODE  *pNode;


    pEntry = pdx->List_SupportedDevices.next;

    // Traverse list to find the desired object
    while (pEntry != &(pdx->List_SupportedDevices))
    {
        // Get the object
        pNode =
            list_entry(
                pEntry,
                PLX_DEVICE_NODE,
                ListEntry
                );

        if ((pNode->bBarsMapped) || (pNode->MapRequestPending))
        {
            DebugPrintf((
                "Release resources for device %p (%04X_%04X [b:%02x s:%02x f:%02x])...\n",
                pNode, pNode->Key.DeviceId, pNode->Key.VendorId,
                pNode->Key.bus, pNode->Key.slot, pNode->Key.function
                ));
        }

        if (pNode->MapRequestPending)
        {
            // Remove any pending mapping requests
            PlxUserMappingRequestFreeAll_ByNode(
                pdx,
                pNode
                );
        }

        if (pNode->bBarsMapped == TRUE)
        {
            // Unmap any kernel mappings to PCI BAR spaces
            PlxPciBarResourcesUnmap(
                pNode,
                (U8)-1          // Specify to unmap all PCI BARs
                );

            // Flag that device has released resources
            pNode->bBarsMapped = FALSE;
        }

        // Jump to next entry
        pEntry = pEntry->next;
    }

    return 0;
}




/*******************************************************************************
 *
 * Function   :  PlxUserMappingRequestFreeAll_ByNode
 *
 * Description:  Removes any pending mapping requests assigned to a specific node
 *
 ******************************************************************************/
VOID
PlxUserMappingRequestFreeAll_ByNode(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_NODE  *pNode
    )
{
    struct list_head *pEntry;
    PLX_USER_MAPPING *pMapObject;


    spin_lock(
        &(pdx->Lock_MapParamsList)
        );

    // Find the mapping parameters from the call to mmap
    pEntry = pdx->List_MapParams.next;

    // Traverse list to find the desired list objects
    while (pEntry != &(pdx->List_MapParams))
    {
        // Get the object
        pMapObject =
            list_entry(
                pEntry,
                PLX_USER_MAPPING,
                ListEntry
                );

        if (pMapObject->pNode == pNode)
        {
            // Decrement request counter
            pNode->MapRequestPending--;

            // Remove the object from the list
            list_del(
                pEntry
                );

            DebugPrintf((
                "Removed request (0x%p) to map to PCI BAR %d\n",
                pMapObject, pMapObject->BarIndex
                ));

            // Release the object
            kfree(
                pMapObject
                );

            // Restart parsing the list from the beginning
            pEntry = pdx->List_MapParams.next;
        }
        else
        {
            // Jump to next item in the list
            pEntry = pEntry->next;
        }
    }

    spin_unlock(
        &(pdx->Lock_MapParamsList)
        );
}




/*******************************************************************************
 *
 * Function   :  PlxDeviceListBuild
 *
 * Description:  Scan the PCI Bus for PCI devices
 *
 ******************************************************************************/
U8
PlxDeviceListBuild(
    DRIVER_OBJECT *pDriverObject
    )
{
    U8                DeviceCount;
    U32               RegValue;
    struct pci_dev   *pPciDev;
    PLX_DEVICE_NODE  *pNode;
    PLX_DEVICE_NODE  *pUpstreamNode;
    DEVICE_EXTENSION *pdx;


    DebugPrintf(("Scanning PCI bus for devices...\n"));

    // Clear device count
    DeviceCount = 0;

    // Initialize upstream node
    pUpstreamNode = NULL;

    // Get device extension
    pdx = pDriverObject->DeviceObject->DeviceExtension;

    // Get OS-generated list of PCI devices
    pPciDev =
        pci_find_device(
            PCI_ANY_ID,
            PCI_ANY_ID,
            NULL
            );

    while (pPciDev)
    {
        DebugPrintf((
            "Adding - %04X %04X  [b:%02x s:%02x f:%02x]\n",
            pPciDev->device, pPciDev->vendor,
            pPciDev->bus->number, PCI_SLOT(pPciDev->devfn),
            PCI_FUNC(pPciDev->devfn)
            ));

        // Increment number of devices found
        DeviceCount++;

        // Allocate memory for new node
        pNode =
            kmalloc(
                sizeof(PLX_DEVICE_NODE),
                GFP_KERNEL
                );

        // Clear node
        RtlZeroMemory(
            pNode,
            sizeof(PLX_DEVICE_NODE)
            );

        // Fill device structure with the PCI information
        pNode->pPciDevice      = pPciDev;
        pNode->Key.bus         = pPciDev->bus->number;
        pNode->Key.slot        = PCI_SLOT(pPciDev->devfn);
        pNode->Key.function    = PCI_FUNC(pPciDev->devfn);
        pNode->Key.VendorId    = pPciDev->vendor;
        pNode->Key.DeviceId    = pPciDev->device;
        pNode->Key.SubVendorId = pPciDev->subsystem_vendor;
        pNode->Key.SubDeviceId = pPciDev->subsystem_device;

        // Save parent device object
        pNode->pdx = pdx;

        // Set PLX chip version
        PlxChipTypeDetect(
            pNode
            );

        // Get PCI header type
        PLX_PCI_REG_READ(
            pNode,
            0x0c,      // PCI Header type / Cache line
            &RegValue
            );

        // Set header type field
        pNode->PciHeaderType = (U8)((RegValue >> 16) & 0x7F);

        // Set port number as unknown
        pNode->PortNumber = (U8)-1;

        // Set default EEPROM width (for 8111/8112 series)
        pNode->Default_EepWidth = 2;

        // Add node to device list
        list_add_tail(
            &(pNode->ListEntry),
            &(pdx->List_SupportedDevices)
            );

        // Check if the all downtream ports are detected
        if (pUpstreamNode != NULL)
        {
            // Support downstream ports for 8000 devices
            if (((pNode->PlxChipType & 0xFF00) == 0x8500) ||
                ((pNode->PlxChipType & 0xFF00) == 0x8600))
            {
                // Detect NT ports and whether virtual or link-side
                if (pNode->PciHeaderType == 0)
                {
                    // Link side upstream port will be itself
                    if (pNode->pUpstreamNode == pUpstreamNode)
                    {
                        pNode->NtPortType = PLX_NT_PORT_LINK;
                    }
                    else
                    {
                        pNode->NtPortType = PLX_NT_PORT_VIRTUAL;
                    }
                }
                else
                {
                    pNode->NtPortType = PLX_NT_PORT_NONE;
                }

                // Differentiate between upstream port and single NT Link-side port
                if (pUpstreamNode->PciHeaderType == 0)
                {
                    // Upstream port is NT link port, clear upstream port
                    pUpstreamNode = NULL;
                }
                else
                {
                    // All downstream nodes are within 1 bus of upstream port
                    if (pNode->Key.bus > (pUpstreamNode->Key.bus + 1))
                    {
                        // Clear upstream port
                        pUpstreamNode = NULL;
                    }
                }
            }
            else
            {
                // Reset upstream port for non-8500 devices
                pUpstreamNode = NULL;
            }
        }

        // For 8000 devices, detect and store parent upstream node
        if (((pNode->PlxChipType & 0xFF00) == 0x8500) ||
            ((pNode->PlxChipType & 0xFF00) == 0x8600) ||
             (pNode->PlxChipType == 0x8114))
        {
            if (pUpstreamNode == NULL)
            {
                // Since this is first node, it must be upstream port, so store it
                pUpstreamNode = pNode;
            }

            // Store the parent for both upstream & downstream ports
            pNode->pUpstreamNode = pUpstreamNode;
        }

        // Jump to next device
        pPciDev =
            pci_find_device(
                PCI_ANY_ID,
                PCI_ANY_ID,
                pPciDev
                );
    }

    DebugPrintf((
        "Device Scan: %d device(s) found\n",
        DeviceCount
        ));

    return DeviceCount;
}




/*******************************************************************************
 *
 * Function   :  GetDeviceNodeFromKey
 *
 * Description:  Traverse device list and return the node associated by key
 *
 ******************************************************************************/
PLX_DEVICE_NODE *
GetDeviceNodeFromKey(
    DEVICE_EXTENSION *pdx,
    PLX_DEVICE_KEY   *pKey
    )
{
    struct list_head *pEntry;
    PLX_DEVICE_NODE  *pNode;


    pEntry = pdx->List_SupportedDevices.next;

    // Traverse list to find the desired object
    while (pEntry != &(pdx->List_SupportedDevices))
    {
        // Get the object
        pNode =
            list_entry(
                pEntry,
                PLX_DEVICE_NODE,
                ListEntry
                );

        if ((pNode->Key.bus      == pKey->bus) &&
            (pNode->Key.slot     == pKey->slot) &&
            (pNode->Key.function == pKey->function))
        {
            return pNode;
        }

        // Jump to next entry
        pEntry = pEntry->next;
    }

    return NULL;
}




/*******************************************************************************
 *
 * Function   :  PlxChipTypeDetect
 *
 * Description:  Attempts to determine PLX chip type and revision
 *
 ******************************************************************************/
BOOLEAN
PlxChipTypeDetect(
    PLX_DEVICE_NODE *pNode
    )
{
    U16 DeviceId;
    U32 RegValue;


    // Reset value
    pNode->PlxChipType = 0;

    // Check PCI E0h for hard-coded ID
    PLX_PCI_REG_READ(
        pNode,
        0xE0,
        &RegValue
        );

    if ((RegValue & 0xFFFF) == PLX_VENDOR_ID)
    {
        pNode->PlxChipType = (U16)(RegValue >> 16);

        // PLX revision should be in PCI E4h
        PLX_PCI_REG_READ(
            pNode,
            0xE4,
            &RegValue
            );

        pNode->PlxRevision = (U8)(RegValue & 0xFF);

        return TRUE;
    }

    // Check PCI 958h for hard-coded ID
    PLX_PCI_REG_READ(
        pNode,
        0x958,
        &RegValue
        );

    if ((RegValue & 0xFFFF) == PLX_VENDOR_ID)
    {
        pNode->PlxChipType = (U16)(RegValue >> 16);

        // PLX revision should be in PCI 95Ch
        PLX_PCI_REG_READ(
            pNode,
            0x95c,
            &RegValue
            );

        pNode->PlxRevision = (U8)(RegValue & 0xFF);

        return TRUE;
    }

    // Get PCI Class code/Revision ID register
    PLX_PCI_REG_READ(
        pNode,
        0x08,           // PCI Revision ID
        &RegValue
        );

    // Default revision to value in chip
    pNode->PlxRevision = (U8)(RegValue & 0xFF);

    // Get current device ID
    DeviceId = pNode->Key.DeviceId;

    // Generalize for PLX 8000 devices
    if (pNode->Key.VendorId == PLX_VENDOR_ID)
    {
        // Don't include 8311 RDK
        if (DeviceId != 0x86e1)
        {
            if (((DeviceId & 0xFF00) == 0x8500) ||
                ((DeviceId & 0xFF00) == 0x8600) ||
                ((DeviceId & 0xFFF0) == 0x8110))
            {
                DeviceId = 0x8000;
            }
        }
    }

    // Hard-coded ID doesn't exist, so use Device/Vendor ID
    switch (((U32)DeviceId << 16) | pNode->Key.VendorId)
    {
        case 0x800010B5:        // All 8000 series devices
            pNode->PlxChipType = pNode->Key.DeviceId;
            break;

        case 0x905010b5:        // 9050/9052
        case 0x520110b5:        // PLX 9052 RDK
            pNode->PlxChipType = 0x9050;
            break;

        case 0x903010b5:        // 9030
        case 0x300110b5:        // PLX 9030 RDK
        case 0x30c110b5:        // PLX 9030 RDK - cPCI
            pNode->PlxChipType = 0x9030;
            break;

        case 0x908010b5:        // 9080
        case 0x040110b5:        // PLX 9080-401B RDK
        case 0x086010b5:        // PLX 9080-860 RDK
            pNode->PlxChipType = 0x9080;
            break;

        case 0x905410b5:        // 9054
        case 0x540610b5:        // PLX 9054 RDK-LITE
        case 0x186010b5:        // PLX 9054-860 RDK
        case 0xc86010b5:        // PLX 9054-860 RDK - cPCI
            pNode->PlxChipType = 0x9054;
            break;

        case 0x905610b5:        // 9056
        case 0x560110b5:        // PLX 9056 RDK-LITE
        case 0x56c210b5:        // PLX 9056-860 RDK
            pNode->PlxChipType = 0x9056;
            break;

        case 0x965610b5:        // 9656
        case 0x960110b5:        // PLX 9656 RDK-LITE
        case 0x96c210b5:        // PLX 9656-860 RDK
            pNode->PlxChipType = 0x9656;
            break;

        case 0x831110b5:        // 8311
        case 0x86e110b5:        // PLX 8311 RDK
            pNode->PlxChipType = 0x8311;
            break;

        case 0x00213388:        // 6140/6152/6254(NT)
            if (pNode->PciHeaderType == 0)
            {
                pNode->PlxChipType = 0x6254;
            }
            else
            {
                // Get 6152 VPD register
                PLX_PCI_REG_READ(
                    pNode,
                    0xA0,
                    &RegValue
                    );

                if ((RegValue & 0xF) == 0x03)    // CAP_ID_VPD
                {
                    pNode->PlxChipType = 0x6152;
                }
                else
                {
                    pNode->PlxChipType = 0x6140;
                }
            }
            break;

        case 0x00223388:        // 6150/6350
        case 0x00a23388:        // 6350
            if (pNode->PlxRevision == 0x20)
            {
                pNode->PlxChipType = 0x6350;
            }
            else
            {
                pNode->PlxChipType = 0x6150;
            }
            break;

        case 0x00263388:        // 6154
            pNode->PlxChipType = 0x6154;
            break;

        case 0x00313388:        // 6156
            pNode->PlxChipType = 0x6156;
            break;

        case 0x00203388:        // 6254
            pNode->PlxChipType = 0x6254;
            break;

        case 0x00303388:        // 6520
        case 0x652010B5:        // 6520
            pNode->PlxChipType = 0x6520;
            break;

        case 0x00283388:        // 6540      - Transparent mode
        case 0x654010B5:        // 6540/6466 - Transparent mode
        case 0x00293388:        // 6540      - Non-transparent mode
        case 0x654110B5:        // 6540/6466 - Non-transparent mode
        case 0x654210B5:        // 6540/6466 - Non-transparent mode
            pNode->PlxChipType = 0x6540;
            break;
    }

    // Detect the PLX chip revision
    PlxChipRevisionDetect(
        pNode
        );

    return TRUE;
}




/*******************************************************************************
 *
 * Function   :  PlxChipRevisionDetect
 *
 * Description:  Attempts to detect the PLX chip revision
 *
 ******************************************************************************/
VOID
PlxChipRevisionDetect(
    PLX_DEVICE_NODE *pNode
    )
{
    U32 RegValue;


    // Get PCI Class code/Revision ID register
    PLX_PCI_REG_READ(
        pNode,
        0x08,           // PCI Revision ID
        &RegValue
        );

    // Default revision to value in chip
    pNode->PlxRevision = (U8)(RegValue & 0xFF);

    // Determine if PLX revision should override default
    switch (pNode->PlxChipType)
    {
        case 0x8111:
            if (pNode->PlxRevision == 0x10)
                pNode->PlxRevision = 0xAA;
            else if (pNode->PlxRevision == 0x20)
                pNode->PlxRevision = 0xBA;
            else if (pNode->PlxRevision == 0x21)
                pNode->PlxRevision = 0xBB;
            break;

        case 0x8112:
            pNode->PlxRevision = 0xAA;
            break;

        case 0x9050:
            if (pNode->PlxRevision == 0x2)
                pNode->PlxRevision = 2;
            else
                pNode->PlxRevision = 1;
            break;

        case 0x9030:
            pNode->PlxRevision = 0xAA;
            break;

        case 0x9080:
            pNode->PlxRevision = 3;
            break;

        case 0x9054:
        case 0x9056:
        case 0x9656:
            // Just use default revision for these chips
            break;

        case 0x8311:
            pNode->PlxRevision = 0xAA;
            break;

        case 0x6140:
            if (pNode->PlxRevision == 0x12)
            {
                pNode->PlxRevision = 0xAA;
            }
            else
            {
                // Revision 0x13 only other
                pNode->PlxRevision = 0xDA;
            }
            break;

        case 0x6150:
            if (pNode->PlxRevision == 0x4)
                pNode->PlxRevision = 0xBB;
            break;

        case 0x6152:
            switch (pNode->PlxRevision)
            {
                case 0x13:
                    pNode->PlxRevision = 0xBA;
                    break;

                case 0x14:
                    pNode->PlxRevision = 0xCA;
                    break;

                case 0x15:
                    pNode->PlxRevision = 0xCC;
                    break;

                case 0x16:
                    pNode->PlxRevision = 0xDA;
                    break;
            }
            break;

        case 0x6154:
            if (pNode->PlxRevision == 0x4)
                pNode->PlxRevision = 0xBB;
            break;

        case 0x6350:
            if (pNode->PlxRevision == 0x20)
                pNode->PlxRevision = 0xAA;
            break;

        case 0x6156:
            if (pNode->PlxRevision == 0x1)
                pNode->PlxRevision = 0xDA;
            break;

        case 0x6254:
            if (pNode->PlxRevision == 0x4)
                pNode->PlxRevision = 0xBB;
            break;

        case 0x6520:
            if (pNode->PlxRevision == 0x2)
                pNode->PlxRevision = 0xBB;
            break;

        case 0x6540:
            if (pNode->PlxRevision == 0x2)
                pNode->PlxRevision = 0xBB;
            break;
    }
}
