/*********************************************************************
 *
 * File Name:
 *
 *      MonCmds.c
 *
 * Description:
 *
 *      Monitor command functions
 *
 ********************************************************************/


#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include "CmdLine.h"
#include "Monitor.h"
#include "MonCmds.h"
#include "PlxApi.h"




/**********************************************************
 *
 * Function   :  Command_Version
 *
 * Description:  Display monitor version information
 *
 *********************************************************/
void
Command_Version(
    void
    )
{
    U8 VersionMajor;
    U8 VersionMinor;
    U8 VersionRevision;

    
    Cons_printf(
        "PLX Console Monitor, v%d.%d%d [%s]\n\n",
        MONITOR_VERSION_MAJOR,
        MONITOR_VERSION_MINOR,
        MONITOR_VERSION_REVISION,
        __DATE__
        );

    // Display PLX API version
    PlxPci_ApiVersion(
        &VersionMajor,
        &VersionMinor,
        &VersionRevision
        );

    Cons_printf(
        "PLX API   : v%d.%d%d",
        VersionMajor & ~(1 << 7),
        VersionMinor,
        VersionRevision
        );

    if (VersionMajor & (1 << 7))
        Cons_printf(" (Demo)\n");
    else
        Cons_printf("\n");

    // Display driver version
    PlxPci_DriverVersion(
        &Gbl_DeviceObj,
        &VersionMajor,
        &VersionMinor,
        &VersionRevision
        );

    Cons_printf(
        "PLX Driver: v%d.%d%d", 
        VersionMajor,
        VersionMinor,
        VersionRevision
        );

    if (pDeviceSelected->PlxChipType != 0)
    {
        Cons_printf(
            " (%04X device)",
            (U16)pDeviceSelected->PlxChipType
            );
    }

    Cons_printf("\n\n");
}




/**********************************************************
 *
 * Function   :  Command_Help
 *
 * Description:  Implements the monitor help command
 *
 *********************************************************/
void
Command_Help(
    void
    )
{
    // Enable throttle output
    ConsoleIoThrottle(TRUE);

    Cons_printf("\n");
    Cons_printf("       -----------  General Commands  -----------\n");
    Cons_printf(" cls       Clear terminal display\n");
    Cons_printf(" screen    Change the screen size (25, 43, or 50 lines)\n");
    Cons_printf(" ver       Display monitor version information\n");
    Cons_printf(" help      Display help screen\n");
    Cons_printf(" exit/quit Exits the application\n");

    Cons_printf("\n");
    Cons_printf("       ------------  Device Access  -------------\n");
    Cons_printf(" dev       Display device list or select new device\n");
    Cons_printf(" set_chip  Force current device to a PLX chip type\n");
    Cons_printf(" pcr       Display PCI Configuration Registers\n");
    Cons_printf(" lcr       Display Local Configuration Registers\n");
    Cons_printf(" rtr       Display Run-time Registers\n");
    Cons_printf(" dma       Display DMA Registers\n");
    Cons_printf(" mqr       Display Message Queue Registers\n");

    Cons_printf("\n");
    Cons_printf("       ------------  EEPROM Access  -------------\n");
    Cons_printf(" eep       Display EEPROM values\n");
    Cons_printf(" eep_load  Load EEPROM values from a file\n");
    Cons_printf(" eep_save  Save EEPROM values to a file\n");

    Cons_printf("\n");
    Cons_printf("       ------------  Memory Access  -------------\n");
    Cons_printf(" db        Read memory  8-bits at a time\n");
    Cons_printf(" dw        Read memory 16-bits at a time\n");
    Cons_printf(" dl        Read memory 32-bits at a time\n");
    Cons_printf(" dq        Read memory 64-bits at a time (if supported)\n");
    Cons_printf(" eb        Write  8-bit data to memory\n");
    Cons_printf(" ew        Write 16-bit data to memory\n");
    Cons_printf(" el        Write 32-bit data to memory\n");
    Cons_printf(" eq        Write 64-bit data to memory (if supported)\n");

    Cons_printf("\n");
    Cons_printf("       -----------  I/O Port Access  ------------\n");
    Cons_printf(" ib        Read I/O port  8-bits at a time\n");
    Cons_printf(" iw        Read I/O port 16-bits at a time\n");
    Cons_printf(" il        Read I/O port 32-bits at a time\n");
    Cons_printf(" ob        Write  8-bit data to I/O port\n");
    Cons_printf(" ow        Write 16-bit data to I/O port\n");
    Cons_printf(" ol        Write 32-bit data to I/O port\n");

    Cons_printf("\n");
    Cons_printf("       --------------  Utilities  ---------------\n");
    Cons_printf(" mf        Interactive memory fill\n");
    Cons_printf(" ms        Interactive memory search\n");
    Cons_printf(" mt8       Test memory in block mode (8-bit)\n");
    Cons_printf(" mt16      Test memory in block mode (16-bit)\n");
    Cons_printf(" mt32      Test memory in block mode (32-bit)\n");

    Cons_printf("\n");
    Cons_printf("       ------------  Miscellaneous  -------------\n");
    Cons_printf(" scan      Scan all PCI buses\n");
    Cons_printf(" vars      Display PCI BAR virtual addresses\n");
    Cons_printf(" portinfo  Display PCI Expres port properties\n");
    Cons_printf(" buff      Display Common buffer properties\n");
    Cons_printf(" reset     Reset the selected device\n");
    Cons_printf(" sleep     Delay for a specified amount of time (in milliseconds)\n");

    Cons_printf("\n");

    // Disable throttle output
    ConsoleIoThrottle(FALSE);
}




/**********************************************************
 *
 * Function   :  Command_Boot
 *
 * Description:  Initiates the system bootstrap
 *
 *********************************************************/
void
Command_Boot(
    void
    )
{
#if !defined(PLX_DOS)
    Cons_printf(
        "Error: The Bootstrap feature is only supported in PlxCm for DOS\n"
        );
#else
    Cons_printf(
        "Error: Bootstrap is not supported directly in PlxCm.\n"
        "       Please use the 16-bit real-mode 'PlxBoot.exe'\n"
        "       DOS application available from PLX Technology.\n"
        );
#endif // PLX_DOS
}




/**********************************************************
 *
 * Function   :  Command_Screen
 *
 * Description:  Changes the screen size
 *
 *********************************************************/
void
Command_Screen(
    char *Args
    )
{
    U16  NumLines;
    char buffer[100];


    // Get new screen size
    GetNextArg(
        &Args,
        buffer
        );

    if (strlen(buffer) == 0)
    {
        // Get current screen height
        NumLines = ConsoleScreenHeightGet();

        Cons_printf("Current screen height is %d lines\n\n", NumLines);

        return;
    }

    // Convert to get new screen height
    NumLines = (U16)atoi(buffer);

    // Set new screen height
    if (ConsoleScreenHeightSet(
            NumLines
            ) == 0)
    {
        Cons_printf("Set new screen height to %d lines\n\n", NumLines);
    }
    else
    {
        Cons_printf("Error: Unable to set new screen height to %d lines\n\n", NumLines);
    }
}




/**********************************************************
 *
 * Function   :  Command_Sleep
 *
 * Description:  Changes the screen size
 *
 *********************************************************/
void
Command_Sleep(
    char *Args
    )
{
    U32  timeout;
    char buffer[100];


    // Get new screen size
    GetNextArg(
        &Args,
        buffer
        );

    if (strlen(buffer) == 0)
    {
        Cons_printf("Useage: sleep {timeout_in_ms}\n\n");
        return;
    }

    // Get timeout value
    timeout = (U32)atol(buffer);

    // Sleep desired amount
    Plx_sleep( timeout );
}




/**********************************************************
 *
 * Function   :  Command_Dev
 *
 * Description:  Selects a new device
 *
 *********************************************************/
void
Command_Dev(
    char *Args
    )
{
    U8           DevNum;
    char         Device[10];
    VOID        *pBuffer;
    DEVICE_NODE *pDeviceNew;


    GetNextArg(
        &Args,
        Device
        );

    if (!strlen(Device))
    {
        DeviceListDisplay();
        Cons_printf("\n");
        return;
    }

    // Convert device number text to number
    DevNum =
        (U8)htol(
                Device
                );

    // Select new device
    pDeviceNew =
        GetDevice(
            DevNum,
            FALSE
            );

    if (pDeviceNew == NULL)
    {
        Cons_printf("Error: Invalid device selection\n\n");
        return;
    }

    // Close current device
    if (pDeviceSelected->bSelected)
    {
        // Unmap PCI BAR spaces
        PciSpacesUnmap();

        // Assign to pointer to avoid compiler error
        pBuffer = &Gbl_PciBufferInfo.UserAddr;

        // Unmap common buffer
        PlxPci_CommonBufferUnmap(
            &Gbl_DeviceObj,
            pBuffer
            );

        Gbl_PciBufferInfo.UserAddr = 0;

        PlxPci_DeviceClose(
            &Gbl_DeviceObj
            );
    }

    // De-select current device
    pDeviceSelected->bSelected = FALSE;

    // Select the new device
    pDeviceSelected = pDeviceNew;

    // Open new device
    PlxPci_DeviceOpen(
        &pDeviceSelected->Key,
        &Gbl_DeviceObj
        );

    // Get virtual addresses for valid PCI spaces
    PciSpacesMap();

    // Get Common buffer properties
    GetCommonBufferProperties();

    // Mark new device as selected
    pDeviceSelected->bSelected = TRUE;

    Cons_printf(
        "Selected: %04x %04x [",
        pDeviceSelected->Key.DeviceId,
        pDeviceSelected->Key.VendorId
        );

    if (pDeviceSelected->PlxChipType != 0)
    {
        if ((pDeviceSelected->PlxChipType == 0x9050) &&
            (pDeviceSelected->PlxRevision == 0x2))
        {
            Cons_printf("9052 - ");
        }
        else
        {
            Cons_printf(
                "%04X - ",
                pDeviceSelected->PlxChipType
                );
        }
    }

    Cons_printf(
        "b:%02x s:%02x f:%02x]\n\n",
        pDeviceSelected->Key.bus,
        pDeviceSelected->Key.slot,
        pDeviceSelected->Key.function
        );
}




/**********************************************************
 *
 * Function   :  Command_SetChip
 *
 * Description:  Sets a device to new chip type
 *
 *********************************************************/
void
Command_SetChip(
    char *Args
    )
{
    char         buffer[20];
    U16          DevNum;
    U32          ChipType;
    PLX_STATUS   rc;
    DEVICE_NODE *pDevice;


    // Get new chip type
    GetNextArg(
        &Args,
        buffer
        );

    if (strlen(buffer) == 0)
    {
        Cons_printf(
            "Usage: set_chip <PlxChipType>\n"
            "\n"
            "       PlxChipType:\n"
            "            0 = reset type & autodetect\n"
            "            Valid PLX 9000, 6000, or 8000 series device\n"
            "            e.g. 9050, 9656, 8111, 6254, 8532, etc\n"
            "\n"
            );

        return;
    }

    // Convert to get chip type
    ChipType = htol(buffer);

    if (ChipType == 0)
    {
        Cons_printf("Resetting chip type...\n");
    }
    else
    {
        Cons_printf(
            "Setting new chip type to: %04X\n",
            ChipType
            );
    }

    // Set new chip type
    rc =
        PlxPci_ChipTypeSet(
            &Gbl_DeviceObj,
            ChipType,
            -1          // Set to autodetect
            );

    if (rc != ApiSuccess)
    {
        Cons_printf(
            "Error: Unable to set new chip type to \"%04X\"\n\n",
            ChipType
            );

        return;
    }

    // Query chip type for updated type
    PlxPci_ChipTypeGet(
        &Gbl_DeviceObj,
        &pDeviceSelected->PlxChipType,
        &pDeviceSelected->PlxRevision
        );

    Cons_printf("Re-selecting device...\n");

    // Get current device number
    DevNum  = 0;
    pDevice = pDeviceList;
    do
    {
        if (pDevice != pDeviceSelected)
        {
            // Jump to next device
            pDevice = pDevice->pNext;
            DevNum++;
        }
    }
    while (pDevice != pDeviceSelected);

    // Build argument
    sprintf(
        buffer,
        "%X",
        DevNum
        );

    // Re-select device
    Command_Dev(buffer);
}




/**********************************************************
 *
 * Function   :  Command_Variables
 *
 * Description:
 *
 *********************************************************/
void
Command_Variables(
    void
    )
{
    U8      i;
    BOOLEAN bVarExists;


    bVarExists = FALSE;

    // Attempt to map all spaces
    for (i=0; i<6; i++)
    {
        if (pDeviceSelected->Va_PciBar[i] != 0)
        {
            if (!bVarExists)
            {
                // Display variables
                Cons_printf(
                    "              Virtual\n"
                    "  Variable    Address        Description\n"
                    " ===============================================\n"
                    );

                bVarExists = TRUE;
            }

            Cons_printf(
                "    V%d       %08lX    BAR %d Virtual Address\n",
                i,
                (unsigned long)pDeviceSelected->Va_PciBar[i],
                i
                );
        }
    }

    if (!bVarExists)
    {
        Cons_printf("Variables:  -- No mapped PCI spaces exist --\n\n");
    }
    else
    {
        Cons_printf(
            "\n"
            "NOTE:\n"
            "   The use of variables in the command-line is not yet\n"
            "   supported.  The BARs are mapped, therefore, the region\n"
            "   is still available by typing the virtual address directly.\n"
            "   For example, if V2=82010000, use \"dl 82010000\".\n\n"
            );
    }
}




/**********************************************************
 *
 * Function   :  Command_DisplayCommonBuffInfo
 *
 * Description:
 *
 *********************************************************/
void
Command_DisplayCommonBuffInfo(
    void
    )
{
    // Only display properties for valid buffer
    if ((Gbl_PciBufferInfo.Size == 0) ||
        (pDeviceSelected->PlxChipType == 0))
    {
        // Check if device is supported
        if (pDeviceSelected->PlxChipType == 0)
        {
            Cons_printf(
                "Host common buffer:  N/A with non-PLX devices\n\n"
                );
        }
        else
        {
            Cons_printf(
                "Host common buffer:  N/A with %04X device\n\n",
                pDeviceSelected->PlxChipType
                );
        }

        return;
    }

    // Display properties
    Cons_printf(
        "Host common buffer (%04X device):\n"
        "  PCI address : %08Xh\n"
        "  Virtual addr: %08Xh\n"
        "  Size        : %08Xh (%d Kb)\n\n",
        pDeviceSelected->PlxChipType,
        (PLX_UINT_PTR)Gbl_PciBufferInfo.PhysicalAddr,
        (PLX_UINT_PTR)Gbl_PciBufferInfo.UserAddr,
        Gbl_PciBufferInfo.Size,
        (Gbl_PciBufferInfo.Size >> 10)
        );
}




/**********************************************************
 *
 * Function   :  Command_Scan
 *
 * Description:
 *
 *********************************************************/
void
Command_Scan(
    void
    )
{
    char        ClassString[100];
    U8          bus;
    U8          slot;
    U8          function;
    U8          DevicesFound;
    U32         RegValue;
    BOOLEAN     bMultiFunction;
    PLX_STATUS  rc;
    DEVICE_NODE TmpNode;


    DevicesFound = 0;

    // Setup temporary device properties
    TmpNode.PlxChipType = 0;
    TmpNode.PlxRevision = 0;

    // Enable throttle output
    ConsoleIoThrottle(TRUE);

    Cons_printf(
        "\n"
        " Bus Slot Fn  Dev  Ven   Device Type\n"
        "===============================================\n"
        );

    for (bus=0; bus < MAX_PCI_BUS; bus++)
    {
        // Check for user abort
        if (Cons_kbhit())
        {
            break;
        }

        for (slot=0; slot < MAX_PCI_DEV; slot++)
        {
            // Check for user abort
            if (Cons_kbhit())
            {
                break;
            }

            function       = 0;
            bMultiFunction = FALSE;

            do
            {
                Cons_printf(
                    "  %02x  %02x  %02x",
                    bus,
                    slot,
                    function
                    );

                RegValue =
                    PlxPci_PciRegisterRead(
                        bus,
                        slot,
                        function,
                        0,         // Device/Vendor ID
                        &rc
                        );

                if ((rc == ApiSuccess) && (RegValue != (U32)-1))
                {
                    Cons_printf(
                        "  %04x %04x",
                        (U16)(RegValue >> 16),
                        (U16)RegValue
                        );

                    // Get device class
                    TmpNode.PciClass =
                        PlxPci_PciRegisterRead(
                            bus,
                            slot,
                            function,
                            0x8,       // Revision ID
                            NULL
                            );

                    // Remove Revision ID
                    TmpNode.PciClass >>= 8;

                    // Get Header type
                    RegValue =
                        PlxPci_PciRegisterRead(
                            bus,
                            slot,
                            function,
                            0xC,       // Header/Cache line size
                            NULL
                            );

                    TmpNode.PciHeaderType = (U8)((RegValue >> 16) & 0x7F);

                    // Check for multi-function device
                    if ((U8)(RegValue >> 16) & (1 << 7))
                    {
                        bMultiFunction = TRUE;

                        // Clear multi-function flag
                        TmpNode.PciHeaderType &= 0x3F;
                    }

                    Device_GetClassString(
                        &TmpNode,
                        ClassString
                        );

                    Cons_printf(
                        "  %s\n",
                        ClassString
                        );

                    DevicesFound++;
                }
                else
                {
                    Cons_printf("\r");
                }

                // Increment function
                function++;
            }
            while (bMultiFunction && (function < MAX_PCI_FUNC));
        }
    }

    Cons_printf(
        "               \n"
        "PCI Bus Scan: "
        );

    if (Cons_kbhit())
    {
        // Clear the character
        Cons_getch();
        Cons_printf("-Aborted-\n\n");
    }
    else
    {
        Cons_printf(
            "%d devices found\n\n",
            DevicesFound
            );
    }

    // Disable throttle output
    ConsoleIoThrottle(FALSE);
}




/**********************************************************
 *
 * Function   :  Command_Reset
 *
 * Description:  Resets a PLX device
 *
 *********************************************************/
void
Command_Reset(
    void
    )
{
    // Check if device is supported
    if (pDeviceSelected->PlxChipType == 0)
    {
        Cons_printf("Error: This command only supports PLX devices\n\n");
        return;
    }

    Cons_printf("Resetting device....");

    PlxPci_DeviceReset(
        &Gbl_DeviceObj
        );

    Cons_printf("Ok\n\n");
}




/**********************************************************
 *
 * Function   :  Command_PciRegs
 *
 * Description:  Implements the PCI register command
 *
 *********************************************************/
void
Command_PciRegs(
    char *Args
    )
{
    U8            i;
    U8            NumArgs;
    U16           Offset;
    U32           Value;
    U32           PlxChip;
    char          buffer[255];
    REGISTER_SET *RegSet;


    // Generalize by chip type
    PlxChip = pDeviceSelected->PlxChipType;

    if (((PlxChip & 0xFF00) == 0x8500) ||
        ((PlxChip & 0xFF00) == 0x8600))
    {
        PlxChip = PlxChip & 0xFF00;
    }

    // Select correct data set
    switch (PlxChip)
    {
        case 0x9030:
        case 0x9056:
        case 0x9656:
        case 0x9054:
        case 0x8311:
            RegSet = Pci9054;
            break;

        case 0x6540:
            RegSet = Pci6540;
            break;

        case 0x8111:
        case 0x8112:
            RegSet = Pci8111;
            break;

        case 0x8500:
        case 0x8600:
            // Check Header Type
            if (pDeviceSelected->PciHeaderType == 1)
            {
                RegSet = Pci_Type_1;
            }
            else
            {
                RegSet = Pci_Type_0;
            }
            break;

        case 0x9050:
        case 0x9080:
        default:
            // Check Header Type
            if (pDeviceSelected->PciHeaderType == 1)
            {
                // Type 1 - P-to-P Bridge
                RegSet = Pci_Type_1;
            }
            else if (pDeviceSelected->PciHeaderType == 2)
            {
                // Type 2 - CardBus Bridge
                RegSet = Pci_Type_2;
            }
            else
            {
                // Type 0 - All other devices
                RegSet = Pci_Type_0;
            }
            break;
    }

    NumArgs = 0;
    Offset  = 0;
    Value   = 0;

    do
    {
        GetNextArg(
            &Args,
            buffer
            );

        i = (U8)strlen(buffer);

        if (i != 0)
        {
            NumArgs++;
            if (NumArgs == 1)
                Offset = (U16)htol(buffer);
            else
                Value = (U32)htol(buffer);
        }
    }
    while (i != 0);

    // Enable throttle output
    if (NumArgs == 0)
        ConsoleIoThrottle(TRUE);

    i = 0;

    while ((RegSet[i].Offset != 0) || (RegSet[i].Description[0] != '\0'))
    {
        if (NumArgs == 0)
        {
            Value =
                PlxPci_PciRegisterReadFast(
                    &Gbl_DeviceObj,
                    RegSet[i].Offset,
                    NULL
                    );

            Cons_printf(
                " %03x: %08x  %s\n",
                RegSet[i].Offset,
                Value,
                RegSet[i].Description
                );
        }
        else
        {
            if (RegSet[i].Offset == Offset)
            {
                if (NumArgs == 1)
                {
                    Value =
                        PlxPci_PciRegisterReadFast(
                            &Gbl_DeviceObj,
                            Offset,
                            NULL
                            );

                    Cons_printf(
                        " %03x: %08x  %s\n",
                        Offset,
                        Value,
                        RegSet[i].Description
                        );

                    return;
                }
                else
                {
                    PlxPci_PciRegisterWriteFast(
                        &Gbl_DeviceObj,
                        Offset,
                        Value
                        );

                    return;
                }
            }
        }                

        i++;
    }

    if (NumArgs == 1)
    {
        Value =
            PlxPci_PciRegisterReadFast(
                &Gbl_DeviceObj,
                Offset,
                NULL
                );

        Cons_printf(
            " %03x: %08x  (unknown register)\n",
            Offset,
            Value
            );
    }
    else if (NumArgs > 1)
    {
        PlxPci_PciRegisterWriteFast(
            &Gbl_DeviceObj,
            Offset,
            Value
            );
    }

    // Disable throttle output
    ConsoleIoThrottle(FALSE);
}




/**********************************************************
 *
 * Function   :  Command_Regs
 *
 * Description:  Implements the register commands
 *
 *********************************************************/
void
Command_Regs(
    U8    Regs,
    char *Args
    )
{
    U8            i;
    U8            NumArgs;
    U32           Offset;
    U32           Value;
    U32           PlxChip;
    char          buffer[255];
    PLX_STATUS    rc;
    REGISTER_SET *RegSet;


    // Check if device is supported
    if (pDeviceSelected->PlxChipType == 0)
    {
        Cons_printf("Error: This command only supports PLX devices\n\n");
        return;
    }

    RegSet = NULL;

    // Generalize by chip type
    PlxChip = pDeviceSelected->PlxChipType;

    if (((PlxChip & 0xFF00) == 0x8500) ||
        ((PlxChip & 0xFF00) == 0x8600))
    {
        PlxChip = PlxChip & 0xFF00;
    }

    // Select desired register set
    switch (PlxChip)
    {
        case 0x9050:
            if (Regs == REGS_LCR)
                RegSet = Lcr9050;
            break;

        case 0x9030:
            if (Regs == REGS_LCR)
                RegSet = Lcr9030;
            break;

        case 0x8111:
        case 0x8112:
            if (Regs == REGS_LCR)
                RegSet = Lcr8111;
            break;

        case 0x8114:
        case 0x8500:
        case 0x8600:
            if (Regs == REGS_LCR)
                RegSet = Lcr8500;
            break;

        case 0x9080:
            switch (Regs)
            {
                case REGS_LCR:
                    RegSet = Lcr9080;
                    break;

                case REGS_RTR:
                    RegSet = Rtr9080;
                    break;

                case REGS_DMA:
                    RegSet = Dma9080;
                    break;

                case REGS_MQR:
                    RegSet = Mqr9080;
                    break;
            }
            break;

        case 0x9056:
        case 0x9656:
        case 0x8311:
            switch (Regs)
            {
                case REGS_LCR:
                    RegSet = Lcr9656;
                    break;

                case REGS_RTR:
                    RegSet = Rtr9080;
                    break;

                case REGS_DMA:
                    RegSet = Dma9054;
                    break;

                case REGS_MQR:
                    RegSet = Mqr9080;
                    break;
            }
            break;

        case 0x9054:
            switch (Regs)
            {
                case REGS_LCR:
                    RegSet = Lcr9054;
                    break;

                case REGS_RTR:
                    RegSet = Rtr9080;
                    break;

                case REGS_DMA:
                    RegSet = Dma9054;
                    break;

                case REGS_MQR:
                    RegSet = Mqr9080;
                    break;
            }
            break;

        case 0x6140:
        case 0x6150:
        case 0x6152:
        case 0x6154:
        case 0x6156:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
            Cons_printf(
                "%04X PCI-to-PCI Bridge does not contain local registers\n\n",
                pDeviceSelected->PlxChipType
                );
            return;

        default:
            Cons_printf(
                "** Local registers not available on %lX devices **\n\n",
                pDeviceSelected->PlxChipType
                );
            return;
    }

    if (RegSet == NULL)
    {
        Cons_printf(
            "Error:  Invalid register set for %lX\n\n",
            pDeviceSelected->PlxChipType
            );
        return;
    }

    NumArgs = 0;
    Offset  = 0;
    Value   = 0;

    do
    {
        GetNextArg(
            &Args,
            buffer
            );

        i = (U8)strlen(buffer);

        if (i != 0)
        {
            NumArgs++;
            if (NumArgs == 1)
                Offset = (U32)htol(buffer);
            else
                Value = (U32)htol(buffer);
        }
    }
    while (i != 0);

    // Enable throttle output
    if (NumArgs == 0)
        ConsoleIoThrottle(TRUE);

    i = 0;

    while ((RegSet[i].Offset != 0) || (RegSet[i].Description[0] != '\0'))
    {
        if (NumArgs == 0)
        {
            Value =
                PlxPci_PlxRegisterRead(
                    &Gbl_DeviceObj,
                    RegSet[i].Offset,
                    &rc
                    );

            Cons_printf(
                " %03x: %08x  %s\n",
                RegSet[i].Offset,
                Value,
                RegSet[i].Description
                );
        }
        else
        {
            if (RegSet[i].Offset == Offset)
            {
                if (NumArgs == 1)
                {
                    Value =
                        PlxPci_PlxRegisterRead(
                            &Gbl_DeviceObj,
                            Offset,
                            &rc
                            );

                    Cons_printf(
                        " %03x: %08x  %s\n",
                        Offset,
                        Value,
                        RegSet[i].Description
                        );

                    return;
                }
                else
                {
                    PlxPci_PlxRegisterWrite(
                        &Gbl_DeviceObj,
                        Offset,
                        Value
                        );

                    return;
                }
            }
        }                

        i++;
    }

    if (NumArgs == 1)
    {
        Value =
            PlxPci_PlxRegisterRead(
                &Gbl_DeviceObj,
                Offset,
                &rc
                );

        Cons_printf(
            " %03x: %08x  (unknown register)\n",
            Offset,
            Value
            );
    }
    else if (NumArgs > 1)
    {
        PlxPci_PlxRegisterWrite(
            &Gbl_DeviceObj,
            Offset,
            Value
            );
    }

    // Disable throttle output
    ConsoleIoThrottle(FALSE);
}




/**********************************************************
 *
 * Function   :  Command_PortProp
 *
 * Description:  Displays port properties
 *
 *********************************************************/
void
Command_PortProp(
    char *Args
    )
{
    PLX_STATUS    status;
    PLX_PORT_PROP PortProp;


    Cons_printf("Port Info: ");
    status =
        PlxPci_GetPortProperties(
            &Gbl_DeviceObj,
            &PortProp
            );

    if (status != ApiSuccess)
    {
        Cons_printf("*ERROR* - API failed (status=%x)\n", status);
        return;
    }
    Cons_printf("\n");

    Cons_printf(
        "    Port Type   : %02d ",
        PortProp.PortType
        );

    switch (PortProp.PortType)
    {
        case PLX_PORT_UNKNOWN:
            Cons_printf("(Unknown?)\n");
            break;

        case PLX_PORT_ENDPOINT:  // PLX_PORT_NON_TRANS
            Cons_printf("(Endpoint or NT port)\n");
            break;

        case PLX_PORT_UPSTREAM:
            Cons_printf("(Upstream port)\n");
            break;

        case PLX_PORT_DOWNSTREAM:
            Cons_printf("(Downstream port)\n");
            break;

        case PLX_PORT_LEGACY_ENDPOINT:
            Cons_printf("(Legacy Endpoint)\n");
            break;

        case PLX_PORT_ROOT_PORT:
            Cons_printf("(Root Port)\n");
            break;

        case PLX_PORT_PCIE_TO_PCI_BRIDGE:
            Cons_printf("(PCIe-to-PCI Bridge)\n");
            break;

        case PLX_PORT_PCI_TO_PCIE_BRIDGE:
            Cons_printf("(PCI-to-PCIe Bridge)\n");
            break;

        case PLX_PORT_ROOT_ENDPOINT:
            Cons_printf("(Root Complex Endpoint)\n");
            break;

        case PLX_PORT_ROOT_EVENT_COLL:
            Cons_printf("(Root Complex Event Collector)\n");
            break;

        default:
            Cons_printf("(N/A)\n");
            break;
    }

    // Port Number
    Cons_printf("    Port Number : ");
    if (PortProp.bNonPcieDevice)
        Cons_printf("--\n");
    else
        Cons_printf("%02d\n", PortProp.PortNumber);

    // Max Payload
    Cons_printf("    Max Payload : ");
    if (PortProp.bNonPcieDevice)
        Cons_printf("--\n");
    else
        Cons_printf("%02d bytes\n", PortProp.MaxPayloadSize);

    // Max Read Req Size
    Cons_printf("    Max Read Req: ");
    if (PortProp.bNonPcieDevice)
        Cons_printf("--\n");
    else
        Cons_printf("%02d bytes\n", PortProp.MaxReadReqSize);

    // Link width
    Cons_printf("    Link Width  : ");
    if (PortProp.bNonPcieDevice)
        Cons_printf("--\n");
    else
        Cons_printf(
            "x%d (Max=x%d)\n",
            PortProp.LinkWidth, PortProp.MaxLinkWidth
            );

    // Link speed
    Cons_printf("    Link Speed  : ");
    if (PortProp.bNonPcieDevice)
        Cons_printf("--\n");
    else
    {
        if (PortProp.LinkSpeed == PLX_LINK_SPEED_5_0_GBPS)
            Cons_printf("5 Gbps (v2.0)");
        else if (PortProp.LinkSpeed == PLX_LINK_SPEED_2_5_GBPS)
            Cons_printf("2.5 Gbps (v1.0)");
        else
            Cons_printf("%d", PortProp.MaxLinkSpeed);

        Cons_printf(" [Max=");
        if (PortProp.MaxLinkSpeed == PLX_LINK_SPEED_5_0_GBPS)
            Cons_printf("5 Gbps (v2.0)]\n");
        else if (PortProp.MaxLinkSpeed == PLX_LINK_SPEED_2_5_GBPS)
            Cons_printf("2.5 Gbps (v1.0)]\n");
        else
            Cons_printf("%d", PortProp.MaxLinkSpeed);
    }
}




/**********************************************************
 *
 * Function   :  Command_Eep
 *
 * Description:  Implements the EEPROM commands
 *
 *********************************************************/
void
Command_Eep(
    char *Args
    )
{
    U8            i;
    U8            NumArgs;
    U16           Offset;
    U16           Value_16;
    U32           Value_32;
    U32           PlxChip;
    char          buffer[255];
    BOOLEAN       b16Bit;
    PLX_STATUS    rc;
    REGISTER_SET *RegSet;


    // Default to 32-bit data
    b16Bit = FALSE;

    // Generalize by chip type
    PlxChip = pDeviceSelected->PlxChipType;

    if (((PlxChip & 0xFF00) == 0x8500) ||
        ((PlxChip & 0xFF00) == 0x8600))
    {
        PlxChip = PlxChip & 0xFF00;
    }

    // Select correct data set
    switch (PlxChip)
    {
        case 0x9050:
            RegSet = Eep9050;
            break;

        case 0x9030:
            RegSet = Eep9030;
            break;

        case 0x9080:
            RegSet = Eep9080;
            break;

        case 0x9054:
            RegSet = Eep9054;
            break;

        case 0x9056:
        case 0x9656:
        case 0x8311:
            RegSet = Eep9656;
            break;

        case 0x6254:
            RegSet = Eep6254;
            b16Bit = TRUE;
            break;

        case 0x6540:
            RegSet = Eep6540;
            b16Bit = TRUE;
            break;

        case 0x8111:
        case 0x8112:
            RegSet = Eep8111;
            break;

        case 0x8114:
        case 0x8500:
        case 0x8600:
            RegSet = Eep8500;
            break;

        case 0:
            Cons_printf("Error: This command only supports PLX devices\n\n");
            return;

        case 0x6140:
        case 0x6150:
        case 0x6152:
        case 0x6154:
        case 0x6156:
        case 0x6350:
        case 0x6520:
        default:
            Cons_printf(
                "** %X EEPROM access not implemented\n\n",
                pDeviceSelected->PlxChipType
                );
            return;
    }

    // Check if an EEPROM is present
    if (PlxPci_EepromPresent(
            &Gbl_DeviceObj,
            &rc
            ) == PLX_EEPROM_STATUS_NONE)
    {
        Cons_printf(
            " -- The PLX chip reports no EEPROM present --\n"
            "\n"
            "Do you want to proceed [y/n]? "
            );

        // Wait for input
        i = (U8)Cons_getch();

        Cons_printf("%c\n\n", i);

        if (i != 'y' && i != 'Y')
        {
            return;
        }
    }

    NumArgs  = 0;
    Offset   = 0;
    Value_32 = 0;

    do
    {
        GetNextArg(
            &Args,
            buffer
            );

        i = (U8)strlen(buffer);

        if (i != 0)
        {
            NumArgs++;
            if (NumArgs == 1)
                Offset = (U16)htol(buffer);
            else
                Value_32 = (U32)htol(buffer);
        }
    }
    while (i != 0);

    if (NumArgs == 0)
    {
        // Generalize by chip type
        PlxChip = pDeviceSelected->PlxChipType;

        if ((PlxChip & 0xFF00) == 0x8600)
        {
            PlxChip = PlxChip & 0xFF00;
        }

        // Process some 8000 devices differently
        switch (PlxChip)
        {
            case 0x8111:
            case 0x8112:
            case 0x8505:
            case 0x8509:
            case 0x8511:
            case 0x8513:
            case 0x8519:
            case 0x8525:
            case 0x8533:
            case 0x8547:
            case 0x8548:
            case 0x8600:
                Command_Eep8000();
                return;
        }
    }

    // Verify offset alignment
    if (NumArgs > 0)
    {
        if (b16Bit)
        {
            if (Offset & (1 << 0))
            {
                Cons_printf("Error: Offset must be aligned on 16-bit boundary\n");
                return;
            }
        }
        else
        {
            if (Offset & (3 << 0))
            {
                Cons_printf("Error: Offset must be aligned on 32-bit boundary\n");
                return;
            }
        }
    }

    // Enable throttle output
    if (NumArgs == 0)
        ConsoleIoThrottle(TRUE);

    i = 0;

    while ((RegSet[i].Offset != 0) || (RegSet[i].Description[0] != '\0'))
    {
        if (NumArgs == 0)
        {
            if (b16Bit)
            {
                PlxPci_EepromReadByOffset_16(
                    &Gbl_DeviceObj,
                    RegSet[i].Offset,
                    &Value_16
                    );

                Cons_printf(
                    " %02x: %04x  %s\n",
                    RegSet[i].Offset,
                    Value_16,
                    RegSet[i].Description
                    );
            }
            else
            {
                PlxPci_EepromReadByOffset(
                    &Gbl_DeviceObj,
                    RegSet[i].Offset,
                    &Value_32
                    );

                Cons_printf(
                    " %03x: %08x  %s\n",
                    RegSet[i].Offset,
                    Value_32,
                    RegSet[i].Description
                    );
            }
        }
        else
        {
            if (RegSet[i].Offset == Offset)
            {
                if (NumArgs == 1)
                {
                    if (b16Bit)
                    {
                        PlxPci_EepromReadByOffset_16(
                            &Gbl_DeviceObj,
                            Offset,
                            &Value_16
                            );

                        Cons_printf(
                            " %02x: %04x  %s\n",
                            Offset,
                            Value_16,
                            RegSet[i].Description
                            );
                    }
                    else
                    {
                        PlxPci_EepromReadByOffset(
                            &Gbl_DeviceObj,
                            Offset,
                            &Value_32
                            );

                        Cons_printf(
                            " %03x: %08x  %s\n",
                            Offset,
                            Value_32,
                            RegSet[i].Description
                            );
                    }

                    return;
                }
                else
                {
                    if (b16Bit)
                    {
                        PlxPci_EepromWriteByOffset_16(
                            &Gbl_DeviceObj,
                            Offset,
                            (U16)Value_32
                            );
                    }
                    else
                    {
                        PlxPci_EepromWriteByOffset(
                            &Gbl_DeviceObj,
                            Offset,
                            Value_32
                            );
                    }

                    return;
                }
            }
        }                

        i++;
    }

    if (NumArgs == 1)
    {
        if (b16Bit)
        {
            PlxPci_EepromReadByOffset_16(
                &Gbl_DeviceObj,
                Offset,
                &Value_16
                );

            Cons_printf(
                " %02x: %04x  (unknown EEPROM location)\n",
                Offset,
                Value_16
                );
        }
        else
        {
            PlxPci_EepromReadByOffset(
                &Gbl_DeviceObj,
                Offset,
                &Value_32
                );

            Cons_printf(
                " %03x: %08x  (unknown EEPROM location)\n",
                Offset,
                Value_32
                );
        }
    }
    else if (NumArgs > 1)
    {
        if (b16Bit)
        {
            PlxPci_EepromWriteByOffset_16(
                &Gbl_DeviceObj,
                Offset,
                (U16)Value_32
                );
        }
        else
        {
            PlxPci_EepromWriteByOffset(
                &Gbl_DeviceObj,
                Offset,
                Value_32
                );
        }
    }

    // Disable throttle output
    ConsoleIoThrottle(FALSE);
}




/**********************************************************
 *
 * Function   :  Command_Eep8000
 *
 * Description:  Implements the EEPROM commands
 *
 *********************************************************/
void
Command_Eep8000(
    void
    )
{
    S16  count;
    S16  ByteCount;
    U16  port;
    U16  Offset;
    U16  Value;
    U16  EepHeader;
    U32  Data;
    U32  PlxChip;
    BOOL bConfigLoadEnabled;
    BOOL bSharedMemLoadEnabled;


    // Setup for correct EEPROM structure processing
    PlxChip = pDeviceSelected->PlxChipType;

    // Enable throttle output
    ConsoleIoThrottle(TRUE);

    Cons_printf(
        "     ------ %04X EEPROM Header ------\n",
        pDeviceSelected->PlxChipType
        );

    // Get EEPROM Header info
    PlxPci_EepromReadByOffset_16(
        &Gbl_DeviceObj,
        0,
        &EepHeader
        );

    bConfigLoadEnabled    = (EepHeader >> 8) & (1 << 0);
    bSharedMemLoadEnabled = (EepHeader >> 8) & (1 << 1);

    PlxPci_EepromReadByOffset_16(
        &Gbl_DeviceObj,
        2,
        &Value
        );

    ByteCount = (U16)Value;

    // Diplay header information
    if ((PlxChip & 0xFF00)== 0x8100)
    {
        Cons_printf(
            " Signature        : %02X (%s)\n"
            " Load Config Regs : %s\n"
            " Load Shared Mem  : %s\n",
            (U8)EepHeader, (U8)EepHeader == 0x5A ? "Valid" : "Invalid",
            bConfigLoadEnabled ? "Yes" : "No",
            bSharedMemLoadEnabled ? "Yes" : "No"
            );
    }
    else
    {
        Cons_printf(
            " Signature : %02X (%s)\n",
            (U8)EepHeader, (U8)EepHeader == 0x5A ? "Valid" : "Invalid"
            );
    }

    Cons_printf(
        "\n"
        "     ----- Registers (%d bytes) -----\n"
        "\n",
        ByteCount
        );

    if (ByteCount != 0)
    {
        Cons_printf(
            "  Port   Offset   Value\n"
            " --------------------------\n"
            );
    }

    // Start at first register
    Offset = 0x4;

    while (ByteCount > 0)
    {
        // Get next register offset
        PlxPci_EepromReadByOffset_16(
            &Gbl_DeviceObj,
            Offset,
            &Value
            );

        // Display port
        if ((PlxChip & 0xFF00) == 0x8100)
        {
            Cons_printf("   --     ");
        }
        else
        {
            // Get bits [15:10] for port number
            port = (Value >> 10) & 0x3F;

            switch (PlxChip)
            {
                case 0x8505:
                case 0x8509:
                case 0x8525:
                case 0x8533:
                case 0x8547:
                case 0x8548:
                    // Port numbers are 1:1 correspondence except for 12 & higher
                    if (port >= 16)
                        port = port - 4;

                    Cons_printf("   %02d     ", port);
                    break;

                case 0x8612:
                case 0x8616:
                case 0x8624:
                case 0x8632:
                case 0x8647:
                case 0x8648:
                    // If [15:10] is 11_1xxxb, then this is NT Link-side
                    if ((port & 0x38) == 0x38)
                    {
                        Cons_printf(" NT Link  ");
                    }
                    else
                    {
                        // Determine the actual port number
                        port = ((port & ~0x7) >> 1) + (port & 0x7);

                        Cons_printf("   %02d     ", port);
                    }
                    break;

                case 0x8511:
                case 0x8513:
                case 0x8519:
                case 0x8604:
                case 0x8606:
                case 0x8608:
                case 0x8614:
                case 0x8618:
                default:
                    // If [15:10] is 11_xxxxb, then this is NT Link-side
                    if ((port & 0x30) == 0x30)
                    {
                        Cons_printf(" NT Link  ");
                    }
                    else
                    {
                        // Port numbers are 1:1 correspondence
                        Cons_printf("   %02d     ", port);
                    }
                    break;
            }

        }

        // Display register offset
        if ((PlxChip & 0xFF00) != 0x8100)
        {
            // Offset DWORD index is [9:0]
            Value = (Value & 0x3FF) << 2;
        }

        Cons_printf("%04X   ", Value);

        // Go to reg data (lower 16-bits)
        Offset += sizeof(U16);

        // Get next register data (lower 16-bits)
        PlxPci_EepromReadByOffset_16(
            &Gbl_DeviceObj,
            Offset,
            &Value
            );

        Data = (U16)Value;

        // Go to reg data (upper 16-bits)
        Offset += sizeof(U16);

        // Get next register data (upper 16-bits)
        PlxPci_EepromReadByOffset_16(
            &Gbl_DeviceObj,
            Offset,
            &Value
            );

        Data |= Value << 16;

        // Update byte count
        ByteCount -= (sizeof(U16) + sizeof(U32));

        Cons_printf("%08X\n", Data);

        // Go to next reg offset
        Offset += sizeof(U16);
    }

    // Display shared memory
    if (((PlxChip & 0xFF00) == 0x8100) && bSharedMemLoadEnabled)
    {
        // Get shared mem size
        PlxPci_EepromReadByOffset_16(
            &Gbl_DeviceObj,
            Offset,
            &Value
            );

        ByteCount = (U16)Value;

        // Go to next data
        Offset += sizeof(U16);

        Cons_printf(
            "\n"
            "     --- Shared Memory (%d bytes) ---",
            ByteCount
            );

        count = 0;

        while (ByteCount > 0)
        {
            // Display start of line
            if ((count & 0xF) == 0)
            {
                Cons_printf(
                    "\n %03X:",
                    count
                    );
            }

            // Get next data
            PlxPci_EepromReadByOffset_16(
                &Gbl_DeviceObj,
                Offset,
                &Value
                );

            Cons_printf(
                " %04X",
                (U16)Value
                );

            // Adjust count
            ByteCount -= sizeof(U16);
            count     += sizeof(U16);

            // Go to next data
            Offset += sizeof(U16);
        }

        Cons_printf("\n");
    }

    Cons_printf("\n");

    // Disable throttle output
    ConsoleIoThrottle(FALSE);
}




/**********************************************************
 *
 * Function   :  Command_EepFile
 *
 * Description:
 *
 *********************************************************/
void
Command_EepFile(
    char    *Args,
    BOOLEAN  bLoadFile
    )
{
    char    buffer[255];
    U8      EepPortSize;
    U16     EepSize;
    U32     PlxChip;
    BOOLEAN bCrc;


    // Get file name argument
    GetNextArg(
        &Args,
        buffer
        );

    if (strlen(buffer) == 0)
    {
        Cons_printf("ERROR: File name not specified\n");
        return;
    }

    // Default to no CRC
    bCrc = FALSE;

    // Default to 32-bit EEPROM port size
    EepPortSize = sizeof(U32);

    // Generalize by chip type
    PlxChip = pDeviceSelected->PlxChipType;

    if ((PlxChip & 0xFF00) == 0x8600)
    {
        PlxChip = PlxChip & 0xFF00;
    }

    // Setup parameters
    switch (PlxChip)
    {
        case 0x8111:
        case 0x8112:
        case 0x8505:
        case 0x8509:
        case 0x8511:
        case 0x8513:
        case 0x8519:
        case 0x8525:
        case 0x8533:
        case 0x8547:
        case 0x8548:
        case 0x8600:
            // Process some 8000 devices differently
            if (bLoadFile)
            {
                Plx8000_EepromFileLoad(
                    buffer      // File name
                    );
            }
            else
            {
                Plx8000_EepromFileSave(
                    buffer,     // File name
                    0           // 0 = Entire EEPROM
                    );
            }
            return;

        case 0x8114:
            if (pDeviceSelected->PlxRevision >= 0xBA)
                EepSize = 0x3EC;
            else
                EepSize = 0x378;
            bCrc = TRUE;
            break;

        case 0x8508:
        case 0x8517:
        case 0x8518:
            EepSize = 0x78F * sizeof(U32);
            bCrc    = TRUE;
            break;

        case 0x8516:
        case 0x8524:
        case 0x8532:
            EepSize = 0xBE4 * sizeof(U32);
            bCrc    = TRUE;
            break;

        case 0:
            Cons_printf("Error: This command only supports PLX devices\n\n");
            return;

        case 0x9050:
        case 0x9056:
        case 0x9656:
        case 0x8311:
            EepSize     = 0x64;
            EepPortSize = sizeof(U16);
            break;

        case 0x9030:
            EepSize     = 0x88;
            EepPortSize = sizeof(U16);
            break;

        case 0x9080:
        case 0x9054:
            EepSize     = 0x58;
            EepPortSize = sizeof(U16);
            break;

        case 0x6140:
        case 0x6150:
        case 0x6152:
        case 0x6154:
        case 0x6156:
        case 0x6254:
        case 0x6350:
        case 0x6520:
        case 0x6540:
        default:
            Cons_printf(
                "** %X EEPROM read/write from/to a file not implemented yet\n\n",
                pDeviceSelected->PlxChipType
                );
            return;
    }

    if (bLoadFile)
    {
        // Load file into EEPROM
        Plx_EepromFileLoad(
            buffer,
            EepSize,
            EepPortSize,
            bCrc
            );
    }
    else
    {
        // Make sure size aligned on 32-bit boundary
        EepSize = (U16)((EepSize + 3) & ~(U32)0x3);

        // Save EEPROM to file
        Plx_EepromFileSave(
            buffer,
            pDeviceSelected->PlxChipType,
            EepSize,
            EepPortSize,
            bCrc
            );
    }
}




/**********************************************************
 *
 * Function   :  Command_MemRead
 *
 * Description:
 *
 *********************************************************/
void
Command_MemRead(
    char *Args,
    char  size
    )
{
    int         i;
    int         x;
    int         CharsToPrint;
    int         SpacesToInsert;
    char        buffer[20];
    char       *CurrArg;
    U8          Repeat;
    U8         *EndAddr;
    U32         CurrValue;
    static U8  *CurrAddr = NULL;
    BOOLEAN     Done;
    BOOLEAN     LineComplete;


    // Enable throttle output
    ConsoleIoThrottle(TRUE);

    do
    {
        i       = 0;
        Repeat  = 0;
        CurrArg = Args;
        EndAddr = CurrAddr + 0x80;

        do
        {
            GetNextArg(
                &CurrArg,
                buffer
                );

            if (!strlen(buffer))
                break;

            if ((buffer[0] == 'r') || (buffer[0] == 'R'))
            {
                Repeat = 1;
            }
            else
            {
                i++;
                if (i == 1)
                {
                    CurrAddr = (U8*)htol(buffer);
                    EndAddr  = CurrAddr + 0x80;
                }
                else
                {
                    EndAddr = CurrAddr + htol(buffer);
                }
            }
        }
        while (CurrArg != NULL);

        i              = 0;
        Done           = FALSE;
        LineComplete   = FALSE;
        CharsToPrint   = 0;
        SpacesToInsert = 0;

        while (!Done)
        {
            if (Cons_kbhit())
            {
                // Clear the character
                Cons_getch();

                Repeat = 0;
                break;
            }

            if (i == 0)
            {
                // Set initial number of characters to display
                CharsToPrint = 0;

                // Set initial Spaces to insert
                SpacesToInsert = (U8)((((size * 2) + 1) * (16 / size)) + 3);

                Cons_printf("%08x: ", (PLX_UINT_PTR)CurrAddr);
            }

            switch (size)
            {
                case sizeof(U8):
                    CurrValue = PlxCm_MemRead_8(CurrAddr);
                    Cons_printf("%02x ", (U8)CurrValue);
                    break;

                case sizeof(U16):
                    CurrValue = PlxCm_MemRead_16(CurrAddr);
                    Cons_printf("%04x ", (U16)CurrValue);
                    break;

                case sizeof(U32):
                    CurrValue = PlxCm_MemRead_32(CurrAddr);
                    Cons_printf("%08x ", CurrValue);
                    break;
            }

            // Store value for printing
            for (x=0; x<size; x++)
            {
                buffer[i+x] = ((U8*)(&CurrValue))[x];

                // Verify printable character
                if ((buffer[i+x] != ' ') &&
                    (!isgraph(buffer[i+x])))
                {
                    buffer[i+x] = '.';
                }
            }

            i        += size;
            CurrAddr += size;

            // Adjust characters to print
            CharsToPrint += size;

            // Adjust spaces
            SpacesToInsert -= (size * 2) + 1;

            if (CurrAddr >= EndAddr)
            {
                Done         = TRUE;
                LineComplete = TRUE;
            }

            if (!LineComplete)
            {
                if (i == 8)
                {
                    SpacesToInsert -= 2;
                    Cons_printf("- ");
                }
                else if (i == 16)
                {
                    i = 0;
                    LineComplete = TRUE;
                }
            }

            if (LineComplete)
            {
                LineComplete = FALSE;

                // Insert necessary spaces
                while (SpacesToInsert--)
                    Cons_printf(" ");

                // Display characters at end of line
                for (x=0; x<CharsToPrint; x++)
                {
                    Cons_printf("%c", buffer[x]);
                }

                Cons_printf("\n");
            }
        }

        if (Repeat)
            Cons_printf("\n");
    }
    while (Repeat);

    // Disable throttle output
    ConsoleIoThrottle(FALSE);
}




/**********************************************************
 *
 * Function   :  Command_MemWrite
 *
 * Description:
 *
 *********************************************************/
void
Command_MemWrite(
    char *Args,
    char  size
    )
{
    U8    i;
    U8    Repeat;
    U8   *Addr;
    U64   Value;
    char  buffer[20];
    char *CurrArg;


    do
    {
        i       = 0;
        Addr    = 0;
        Repeat  = 0;
        CurrArg = Args;

        do
        {
            GetNextArg(
                &CurrArg,
                buffer
                );

            if (!strlen(buffer))
            {
                if (i <= 1)
                    Cons_printf("??? Error: missing argument(s)\n");
                break;
            }

            if (!strcmp(buffer, "r") || !strcmp(buffer, "R"))
                Repeat = 1;
            else
            {
                i++;
                if (i == 1)
                    Addr = (U8*)htol(buffer);
                else
                {
                    Value = htol(buffer);
                    switch (size)
                    {
                        case sizeof(U8):
                            PlxCm_MemWrite_8(Addr, (U8)Value);
                            break;

                        case sizeof(U16):
                            PlxCm_MemWrite_16(Addr, (U16)Value);
                            break;

                        case sizeof(U32):
                            PlxCm_MemWrite_32(Addr, (U32)Value);
                            break;
                    }

                    Addr += size;
                }
            }
        }
        while (CurrArg != NULL);
    }
    while (Repeat);
}




/**********************************************************
 *
 * Function   :  Command_IoRead
 *
 * Description:
 *
 *********************************************************/
void
Command_IoRead(
    char *Args,
    char  size
    )
{
    int         i;
    int         x;
    int         CharsToPrint;
    int         SpacesToInsert;
    char        buffer[20];
    char       *CurrArg;
    U8          Repeat;
    U32         EndAddr;
    U32         CurrValue;
    static U32  CurrAddr = 0;
    BOOLEAN     Done;
    BOOLEAN     LineComplete;


    // Enable throttle output
    ConsoleIoThrottle(TRUE);

    do
    {
        i       = 0;
        Repeat  = 0;
        CurrArg = Args;
        EndAddr = CurrAddr + 0x80;

        do
        {
            GetNextArg(
                &CurrArg,
                buffer
                );

            if (!strlen(buffer))
                break;

            if ((buffer[0] == 'r') || (buffer[0] == 'R'))
            {
                Repeat = 1;
            }
            else
            {
                i++;
                if (i == 1)
                {
                    CurrAddr = (U32)htol(buffer);
                    EndAddr  = CurrAddr + 0x80;
                }
                else
                {
                    EndAddr = CurrAddr + (U32)htol(buffer);
                }
            }
        }
        while (CurrArg != NULL);

        i              = 0;
        Done           = FALSE;
        LineComplete   = FALSE;
        CharsToPrint   = 0;
        SpacesToInsert = 0;

        while (!Done)
        {
            if (Cons_kbhit())
            {
                // Clear the character
                Cons_getch();

                Repeat = 0;
                break;
            }

            if (i == 0)
            {
                // Set initial number of characters to display
                CharsToPrint = 0;

                // Set initial Spaces to insert
                SpacesToInsert = (U8)((((size * 2) + 1) * (16 / size)) + 3);

                Cons_printf(
                    "%08x: ",
                    CurrAddr
                    );
            }

            switch (size)
            {
                case sizeof(U8):
                    PlxPci_IoPortRead(
                        &Gbl_DeviceObj,
                        CurrAddr,
                        &CurrValue,
                        sizeof(U8),
                        BitSize8
                        );

                    Cons_printf(
                        "%02x ",
                        (U8)CurrValue
                        );
                    break;

                case sizeof(U16):
                    PlxPci_IoPortRead(
                        &Gbl_DeviceObj,
                        CurrAddr,
                        &CurrValue,
                        sizeof(U16),
                        BitSize16
                        );

                    Cons_printf(
                        "%04x ",
                        (U16)CurrValue
                        );
                    break;

                case sizeof(U32):
                    PlxPci_IoPortRead(
                        &Gbl_DeviceObj,
                        CurrAddr,
                        &CurrValue,
                        sizeof(U32),
                        BitSize32
                        );

                    Cons_printf(
                        "%08x ",
                        CurrValue
                        );
                    break;
            }

            // Store value for printing
            for (x=0; x<size; x++)
            {
                buffer[i+x] = ((U8*)(&CurrValue))[x];

                // Verify printable character
                if ((buffer[i+x] != ' ') &&
                    (!isgraph(buffer[i+x])))
                {
                    buffer[i+x] = '.';
                }
            }

            i        += size;
            CurrAddr += size;

            // Adjust characters to print
            CharsToPrint += size;

            // Adjust spaces
            SpacesToInsert -= (size * 2) + 1;

            if (CurrAddr >= EndAddr)
            {
                Done         = TRUE;
                LineComplete = TRUE;
            }

            if (!LineComplete)
            {
                if (i == 8)
                {
                    SpacesToInsert -= 2;
                    Cons_printf("- ");
                }
                else if (i == 16)
                {
                    i = 0;
                    LineComplete = TRUE;
                }
            }

            if (LineComplete)
            {
                LineComplete = FALSE;

                // Insert necessary spaces
                while (SpacesToInsert--)
                    Cons_printf(" ");

                // Display characters at end of line
                for (x=0; x<CharsToPrint; x++)
                {
                    Cons_printf("%c", buffer[x]);
                }

                Cons_printf("\n");
            }
        }

        if (Repeat)
            Cons_printf("\n");
    }
    while (Repeat);

    // Disable throttle output
    ConsoleIoThrottle(FALSE);
}




/**********************************************************
 *
 * Function   :  Command_IoWrite
 *
 * Description:
 *
 *********************************************************/
void
Command_IoWrite(
    char *Args,
    char  size
    )
{
    U8    i;
    U8    Repeat;
    U32   Addr;
    char  buffer[20];
    char *CurrArg;
    U32   Value;


    do
    {
        i       = 0;
        Addr    = 0;
        Repeat  = 0;
        CurrArg = Args;

        do
        {
            GetNextArg(
                &CurrArg,
                buffer
                );

            if (!strlen(buffer))
            {
                if (i <= 1)
                    Cons_printf("??? Error: missing argument(s)\n");
                break;
            }

            if (!strcmp(buffer, "r") || !strcmp(buffer, "R"))
                Repeat = 1;
            else
            {
                i++;
                if (i == 1)
                    Addr = (U32)htol(buffer);
                else
                {
                    Value = (U32)htol(buffer);
                    switch (size)
                    {
                        case sizeof(U8):
                            PlxPci_IoPortWrite(
                                &Gbl_DeviceObj,
                                Addr,
                                &Value,
                                sizeof(U8),
                                BitSize8
                                );
                            break;

                        case sizeof(U16):
                            PlxPci_IoPortWrite(
                                &Gbl_DeviceObj,
                                Addr,
                                &Value,
                                sizeof(U16),
                                BitSize16
                                );
                            break;

                        case sizeof(U32):
                            PlxPci_IoPortWrite(
                                &Gbl_DeviceObj,
                                Addr,
                                &Value,
                                sizeof(U32),
                                BitSize32
                                );
                            break;
                    }

                    Addr += size;
                }
            }
        }
        while (CurrArg != NULL);
    }
    while (Repeat);
}
