/******************************************************************************
 *
 * File Name:
 *
 *      Monitor.c
 *
 * Description:
 *
 *      Monitor main module
 *
 ******************************************************************************/


#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "CmdLine.h"
#include "Monitor.h"
#include "MonCmds.h"
#include "PlxApi.h"
#include "RegDefs.h"




/**********************************************
 *                 Globals
 **********************************************/
char               CmdHistory[MAX_CMD_HISTORY][MAX_BUFFER_SZ];
DEVICE_NODE       *pDeviceSelected;
PLX_PHYSICAL_MEM   Gbl_PciBufferInfo;
PLX_DEVICE_OBJECT  Gbl_DeviceObj;




/**********************************************************
 *
 * Function   :  main
 *
 * Description:
 *
 *********************************************************/
int
main(
    int   argc,
    char *argv[]
    )
{
    int   ExitCode;
    VOID *pBuffer;


    // Initialize the console
    ConsoleInitialize();

    // Set default exit code
    ExitCode = 0;

    // Check if Windows NT flavor is running
    if (PlxPreventInWindows())
    {
        Cons_printf(
            "\n"
            "      ERROR:\n"
            "\n"
            "        This application is designed for execution in a\n"
            "        true DOS environment.  The current OS detected is\n"
            "        Windows NT/2000/XP.  This application will not\n"
            "        function properly in this environment.\n"
            "\n"
            "                   Press any key to exit..."
            );

        Cons_getch();
        Cons_printf("\n");
        ExitCode = 1;
        goto __Exit_App;
    }

    // Make sure no device is selected
    pDeviceSelected = NULL;

    // Clear DMA buffer information
    Gbl_PciBufferInfo.UserAddr     = 0;
    Gbl_PciBufferInfo.PhysicalAddr = 0;
    Gbl_PciBufferInfo.Size         = 0;

    // Build device list
    DeviceListCreate();

    // Start the monitor
    Monitor();

    // Unmap PCI BAR spaces
    PciSpacesUnmap();

    // Assign to pointer to avoid compiler error
    pBuffer = &Gbl_PciBufferInfo.UserAddr;

    // Unmap common buffer
    PlxPci_CommonBufferUnmap(
        &Gbl_DeviceObj,
        pBuffer
        );

    // Make sure device is closed
    PlxPci_DeviceClose(
        &Gbl_DeviceObj
        );

__Exit_App:
    // Free device list
    DeviceListFree();

    // Restore console
    ConsoleEnd();

    exit(ExitCode);
}




/**********************************************************
 *
 * Function   :  Monitor
 *
 * Description:
 *
 *********************************************************/
void
Monitor(
    void
    )
{
    U8      i;
    int     CurrCmd;
    int     FirstCmd;
    int     LastCmd;
    int     NextInp;
    char    buffer[MAX_BUFFER_SZ];
    BOOLEAN bFlag;


    // Select first PLX device
    pDeviceSelected =
        GetDevice(
            0,
            TRUE
            );

    if (pDeviceSelected == NULL)
    {
        Cons_printf(
            "\n"
            "             Error: ** No PCI Devices found **\n"
            "\n"
            "                  Press any key to exit..."
            );

        // Wait for keypress
        Cons_getch();

        Cons_printf("\n");

        return;
    }

    // Select the device
    PlxPci_DeviceOpen(
        &pDeviceSelected->Key,
        &Gbl_DeviceObj
        );

    pDeviceSelected->bSelected = TRUE;

    // Get virtual addresses for valid PCI spaces
    PciSpacesMap();

    // Get Common buffer properties
    GetCommonBufferProperties();

    Cons_clear();

    // Display devices
    DeviceListDisplay();

    Cons_printf(
        "\n"
        "PLX Console Monitor, v%d.%d%d [%s]\n"
        "Copyright (c) PLX Technology, Inc.\n\n",
        MONITOR_VERSION_MAJOR,
        MONITOR_VERSION_MINOR,
        MONITOR_VERSION_REVISION,
        __DATE__
        );

    // Initialize variables
    i           = 0;
    CurrCmd     = 0;
    FirstCmd    = 0;
    LastCmd     = 0;

    memset(
        CmdHistory,
        '\0',
        sizeof(CmdHistory)
        );

    memset(
        buffer,
        '\0',
        MAX_BUFFER_SZ
        );

    Cons_printf(MONITOR_PROMPT);

    do
    {
        // Wait for keypress
        NextInp = Cons_getch();

        switch (NextInp)
        {
            case '\n':
            case '\r':
                Cons_printf("\n");

                if (i != 0)
                {
                    strcpy(
                        CmdHistory[FirstCmd],
                        buffer
                        );

                    if (ProcessCommand(
                            CmdHistory[FirstCmd]
                            ) == FALSE)
                    {
                        // Exit selected
                        return;
                    }

                    FirstCmd = (FirstCmd + 1) % MAX_CMD_HISTORY;
                    CurrCmd  = FirstCmd;
                    if (LastCmd == FirstCmd)
                        LastCmd = (LastCmd + 1) % MAX_CMD_HISTORY;

                    CmdHistory[FirstCmd][0] = '\0';

                    memset(
                        buffer,
                        '\0',
                        i
                        );

                    i = 0;
                }

                Cons_printf(MONITOR_PROMPT);
                break;

            case '\b':                // BackSpace
            case 127:
                if (i != 0)
                {
                    i--;
                    buffer[i] = '\0';
                    Cons_printf("\b \b");
                }
                break;

            case '\t':
                // Ignore TAB
                break;

            case 27:                  // ESC

                // Delay for a bit to wait for additional keys
                Plx_sleep(100);

                // Process ESC only of no other key is pressed
                if (Cons_kbhit() == FALSE)
                {
                    // Clear the command-line
                    while (i)
                    {
                        Cons_printf("\b \b");
                        i--;
                        buffer[i] = '\0';
                    }
                }
                break;

            case 0x0:               // Extended key (DOS)
            case 0xe0:              // Extended key (Windows)
            case '[':               // Extended key (Linux)

                bFlag = TRUE;

                // Get extended code
                switch (Cons_getch())
                {
                    case 'A':            // Up arrow (Linux)
                    case 'H':            // Up arrow (Windows)
                        if (CurrCmd != LastCmd)
                        {
                            if (CurrCmd == 0)
                                CurrCmd = MAX_CMD_HISTORY - 1;
                            else
                                CurrCmd = CurrCmd - 1;
                        }
                        break;

                    case 'B':           // Down arrow (Linux)
                    case 'P':           // Down arrow (Windows)
                        if (CurrCmd != FirstCmd)
                        {
                            if (CurrCmd == MAX_CMD_HISTORY - 1)
                                CurrCmd = 0;
                            else
                                CurrCmd = CurrCmd + 1;
                        }
                        break;

                    case 'D':           // Left arrow (Linux)
                    case 'K':           // Left arrow (Windows)
                        // Not currently handled
                        bFlag = FALSE;
                        break;

                    case 'C':           // Right arrow (Linux)
                    case 'M':           // Right arrow (Windows)
                        // Not currently handled
                        bFlag = FALSE;
                        break;

                    default:
                        // Unsupported key, do nothing
                        bFlag = FALSE;
                        break;
                }

                if (bFlag)
                {
                    // Clear the command-line
                    while (i)
                    {
                        Cons_printf("\b \b");
                        i--;
                        buffer[i] = '\0';
                    }

                    // Copy command from history
                    strcpy(
                        buffer,
                        CmdHistory[CurrCmd]
                        );

                    // Display new command
                    Cons_printf(
                        buffer
                        );

                    i = (U8)strlen(buffer);
                }
                break;

            default:
                // Check for overflow
                if (i < (MAX_BUFFER_SZ - 1))
                {
                    buffer[i] = (char)NextInp;
                    Cons_printf("%c", (char)NextInp);
                    i++;
                }
                break;
        }
    }
    while (1);
}




/**********************************************************
 *
 * Function   :  ProcessCommand
 *
 * Description:
 *
 *********************************************************/
BOOLEAN
ProcessCommand(
    char *buffer
    )
{
    U8   i;
    char Cmd[MAX_CMD_SZ];


    GetNextArg(
        &buffer,
        Cmd
        );

    // Convert command to lower-case
    i=0;
    while (Cmd[i] != '\0')
    {
        Cmd[i] = (char)tolower(Cmd[i]);
        i++;
    }

    if (!strcmp(Cmd, "cls") || !strcmp(Cmd, "clear"))
    {
        Cons_clear();
    }
    else if (!strncmp(Cmd, "//", 2))
    {
        // User entered a comment, ignore it
    }
    else if (!strcmp(Cmd, "screen"))
    {
        Command_Screen(
            buffer
            );
    }
    else if (!strcmp(Cmd, "exit") ||
             !strcmp(Cmd, "quit") ||
             !strcmp(Cmd, "q"))
    {
        return FALSE;
    }
    else if (!strcmp(Cmd, "ver"))
    {
        Command_Version();
    }
    else if (!strcmp(Cmd, "?") || !strcmp(Cmd, "help"))
    {
        Command_Help();
    }
    else if (!strcmp(Cmd, "dev"))
    {
        Command_Dev(
            buffer
            );
    }
    else if (!strcmp(Cmd, "sleep"))
    {
        Command_Sleep(
            buffer
            );
    }
    else if (!strcmp(Cmd, "boot"))
    {
        Command_Boot();
    }
    else if (!strcmp(Cmd, "set_chip"))
    {
        Command_SetChip(
            buffer
            );
    }
    else if (!strcmp(Cmd, "vars"))
    {
        Command_Variables();
    }
    else if (!strcmp(Cmd, "buff"))
    {
        Command_DisplayCommonBuffInfo();
    }
    else if (!strcmp(Cmd, "scan"))
    {
        Command_Scan();
    }
    else if (!strcmp(Cmd, "reset"))
    {
        Command_Reset();
    }
    else if (!strcmp(Cmd, "pcr"))
    {
        Command_PciRegs(
            buffer
            );
    }
    else if (!strcmp(Cmd, "lcr"))
    {
        Command_Regs(
            REGS_LCR,
            buffer
            );
    }
    else if (!strcmp(Cmd, "rtr"))
    {
        Command_Regs(
            REGS_RTR,
            buffer
            );
    }
    else if (!strcmp(Cmd, "dma"))
    {
        Command_Regs(
            REGS_DMA,
            buffer
            );
    }
    else if (!strcmp(Cmd, "mqr"))
    {
        Command_Regs(
            REGS_MQR,
            buffer
            );
    }
    else if (!strcmp(Cmd, "portinfo"))
    {
        Command_PortProp(
            buffer
            );
    }
    else if (!strcmp(Cmd, "eep"))
    {
        Command_Eep(
            buffer
            );
    }
    else if (!strcmp(Cmd, "eep_load"))
    {
        Command_EepFile(
            buffer,
            TRUE        // Load from file
            );
    }
    else if (!strcmp(Cmd, "eep_save"))
    {
        Command_EepFile(
            buffer,
            FALSE       // Save to file
            );
    }
    else if (!strcmp(Cmd, "db"))
    {
        Command_MemRead(
            buffer,
            sizeof(U8)
            );
    }
    else if (!strcmp(Cmd, "dw"))
    {
        Command_MemRead(
            buffer,
            sizeof(U16)
            );
    }
    else if (!strcmp(Cmd, "dl"))
    {
        Command_MemRead(
            buffer,
            sizeof(U32)
            );
    }
    else if (!strcmp(Cmd, "eb"))
    {
        Command_MemWrite(
            buffer,
            sizeof(U8)
            );
    }
    else if (!strcmp(Cmd, "ew"))
    {
        Command_MemWrite(
            buffer,
            sizeof(U16)
            );
    }
    else if (!strcmp(Cmd, "el"))
    {
        Command_MemWrite(
            buffer,
            sizeof(U32)
            );
    }
    else if (!strcmp(Cmd, "ib"))
    {
        Command_IoRead(
            buffer,
            sizeof(U8)
            );
    }
    else if (!strcmp(Cmd, "iw"))
    {
        Command_IoRead(
            buffer,
            sizeof(U16)
            );
    }
    else if (!strcmp(Cmd, "il"))
    {
        Command_IoRead(
            buffer,
            sizeof(U32)
            );
    }
    else if (!strcmp(Cmd, "ob"))
    {
        Command_IoWrite(
            buffer,
            sizeof(U8)
            );
    }
    else if (!strcmp(Cmd, "ow"))
    {
        Command_IoWrite(
            buffer,
            sizeof(U16)
            );
    }
    else if (!strcmp(Cmd, "ol"))
    {
        Command_IoWrite(
            buffer,
            sizeof(U32)
            );
    }
    else if (!strcmp(Cmd, "mf"))
    {
        Cons_printf(" ** Not implemented yet **\n\n");
    }
    else if (!strcmp(Cmd, "ms"))
    {
        Cons_printf(" ** Not implemented yet **\n\n");
    }
    else
    {
        Cons_printf(
            "\"%s\" is not a valid command. Type "
            "\"help\" for more information.\n\n",
            buffer-strlen(Cmd)
            );
    }

    return TRUE;
}




/**********************************************************
 *
 * Function   :  PciSpacesMap
 *
 * Description:
 *
 *********************************************************/
void
PciSpacesMap(
    void
    )
{
    U8 i;


    for (i=0; i<6; i++)
    {
        // Map PCI BAR into user space
        PlxPci_PciBarMap(
            &Gbl_DeviceObj,
            i,
            (VOID**)&(pDeviceSelected->Va_PciBar[i])
            );
    }
}




/**********************************************************
 *
 * Function   :  PciSpacesUnmap
 *
 * Description:
 *
 *********************************************************/
void
PciSpacesUnmap(
    void
    )
{
    U8 i;


    // Skip if no device is selected
    if (pDeviceSelected == NULL)
        return;

    for (i=0; i<6; i++)
    {
        // Unmap PCI BAR from user space
        PlxPci_PciBarUnmap(
            &Gbl_DeviceObj,
            (VOID**)&(pDeviceSelected->Va_PciBar[i])
            );
    }
}




/**********************************************************
 *
 * Function   :  GetCommonBufferProperties
 *
 * Description:
 *
 *********************************************************/
void
GetCommonBufferProperties(
    void
    )
{
    VOID *pBuffer;


    // Get PCI buffer properties
    PlxPci_CommonBufferProperties(
        &Gbl_DeviceObj,
        &Gbl_PciBufferInfo
        );

    // Map buffer into user space
    PlxPci_CommonBufferMap(
        &Gbl_DeviceObj,
        &pBuffer
        );

    // Assign to pointer to avoid compiler error
    Gbl_PciBufferInfo.UserAddr = (PLX_UINT_PTR)pBuffer;
}
