/******************************************************************************
 *
 * File Name:
 *
 *      PlxDma.c
 *
 * Description:
 *
 *      Demonstrates using the PLX API to perform DMA transfers
 *
 * Revision History:
 *
 *      12-01-07 : PLX SDK v5.20
 *
 ******************************************************************************/


#include "PlxApi.h"

#if defined(PLX_MSWINDOWS)
    #include "..\\Shared\\ConsFunc.h"
    #include "..\\Shared\\PlxInit.h"
#endif

#if defined(PLX_LINUX)
    #include "ConsFunc.h"
    #include "PlxInit.h"
#endif




/**********************************************
 *               Functions
 *********************************************/
void
PerformDma(
    PLX_DEVICE_OBJECT *pDevice
    );




/******************************************************************************
 *
 * Function   :  main
 *
 * Description:  The main entry point
 *
 *****************************************************************************/
int
main(
    void
    )
{
    S8                DeviceSelected;
    U8                PlxChipRevision;
    U32               PlxChipType;
    PLX_STATUS        rc;
    PLX_DEVICE_KEY    DeviceKey;
    PLX_DEVICE_OBJECT Device;


    ConsoleInitialize();

    Cons_clear();

    Cons_printf(
        "\n\n"
        "\t\t         PLX DMA Sample Application\n"
        "\t\t                October 2007\n\n"
        );


    /************************************
    *         Select Device
    ************************************/
    DeviceSelected =
        SelectDevice(
            &DeviceKey
            );

    if (DeviceSelected == -1)
    {
        ConsoleEnd();
        exit(0);
    }

    rc =
        PlxPci_DeviceOpen(
            &DeviceKey,
            &Device
            );

    if (rc != ApiSuccess)
    {
        Cons_printf("\n   ERROR: Unable to find or select a PLX device\n");
        PlxSdkErrorDisplay(rc);
        _Pause;
        ConsoleEnd();
        exit(-1);
    }

    Cons_clear();

    Cons_printf(
        "\nSelected: %04x %04x [b:%02x  s:%02x  f:%02x]\n\n",
        DeviceKey.DeviceId, DeviceKey.VendorId,
        DeviceKey.bus, DeviceKey.slot, DeviceKey.function
        );

    // Get PLX Chip Type
    PlxPci_ChipTypeGet(
        &Device,
        &PlxChipType,
        &PlxChipRevision
        );

    // Verify chip is supported
    switch (PlxChipType)
    {
        case 0x9080:
        case 0x9054:
        case 0x9056:
        case 0x9656:
        case 0x8311:
            break;

        default:
            Cons_printf(
                "ERROR: DMA not supported by the selected device (%04X)\n",
                PlxChipType
                );
            goto _Exit_App;
    }


    /************************************
     *        Perform the DMA
     ************************************/
    PerformDma(
        &Device
        );


    /************************************
     *        Close the Device
     ***********************************/
_Exit_App:
    PlxPci_DeviceClose(
        &Device
        );

    _Pause;

    Cons_printf("\n\n");

    ConsoleEnd();

    exit(0);
}




/********************************************************
 *
 *******************************************************/
void
PerformDma(
    PLX_DEVICE_OBJECT *pDevice
    )
{
    U8                DmaChannel;
    U8               *pBuffer = NULL;
    U16               ChannelInput;
    U32               LocalAddress;
    PLX_STATUS        rc;
    PLX_DMA_PROP      DmaProp;
    PLX_INTERRUPT     PlxInterrupt;
    PLX_DMA_PARAMS    DmaParams;
    PLX_PHYSICAL_MEM  PciBuffer;
    PLX_NOTIFY_OBJECT NotifyObject;


    Cons_printf(
        "Description:\n"
        "     This sample demonstrates usage of the PLX DMA API by\n"
        "     performing DMA transfers and waiting for DMA completion.\n"
        );

    Cons_printf(
        "\n"
        " WARNING: There is no safeguard mechanism to protect against invalid\n"
        "          local bus addresses.  Please be careful when selecting local\n"
        "          addresses to transfer data to/from.  The DMA engine will hang\n"
        "          if an invalid address is accessed.\n"
        "\n\n"
        );

    Cons_printf("Please enter a valid local address ----> ");
    Cons_scanf("%x", &LocalAddress);

    Cons_printf("Please select a DMA channel (0 or 1) --> ");
    Cons_scanf("%hd", &ChannelInput);
    Cons_printf("\n");

    if ((ChannelInput != 0) && (ChannelInput != 1))
    {
        Cons_printf("ERROR: Unsupported DMA channel, test aborted\n");
        return;
    }

    DmaChannel = (U8)ChannelInput;

    // Get DMA buffer parameters
    PlxPci_CommonBufferProperties(
        pDevice,
        &PciBuffer
        );


    // Clear DMA structure
    memset(&DmaProp, 0, sizeof(PLX_DMA_PROP));

    // Initialize the DMA channel
    DmaProp.LocalBusWidth = 3;   // 32-bit
    DmaProp.ReadyInput    = 1;

    Cons_printf("  Open Channel %i for DMA......... ", DmaChannel);
    rc =
        PlxPci_DmaChannelOpen(
            pDevice,
            DmaChannel,
            &DmaProp
            );

    if (rc == ApiSuccess)
    {
        Cons_printf("Ok\n");
    }
    else
    {
        Cons_printf("*ERROR* - API failed\n");
        PlxSdkErrorDisplay(rc);
    }


    Cons_printf("  Register for notification...... ");

    // Clear interrupt fields
    memset(&PlxInterrupt, 0, sizeof(PLX_INTERRUPT));

    // Setup to wait for either DMA channel
    PlxInterrupt.DmaChannel_0 = 1;
    PlxInterrupt.DmaChannel_1 = 1;

    rc =
        PlxPci_NotificationRegisterFor(
            pDevice,
            &PlxInterrupt,
            &NotifyObject
            );

    if (rc != ApiSuccess)
    {
        Cons_printf("*ERROR* - API failed\n");
        PlxSdkErrorDisplay(rc);
    }
    else
    {
        Cons_printf(
            "Ok\n"
            );
    }


    // Transfer the Data
    Cons_printf("  Perform Block DMA transfer..... ");

    // Clear DMA data
    memset(&DmaParams, 0, sizeof(PLX_DMA_PARAMS));

    DmaParams.u.PciAddrLow  = (U32)PciBuffer.PhysicalAddr;
    DmaParams.LocalAddr     = LocalAddress;
    DmaParams.ByteCount     = 0x10;
    DmaParams.LocalToPciDma = 1;

    rc =
        PlxPci_DmaTransferBlock(
            pDevice,
            DmaChannel,
            &DmaParams,
            0          // Don't wait for completion
            );

    if (rc == ApiSuccess)
    {
        Cons_printf("Ok\n");

        Cons_printf("  Wait for interrupt event....... ");

        rc =
            PlxPci_NotificationWait(
                pDevice,
                &NotifyObject,
                5 * 1000
                );

        switch (rc)
        {
            case ApiSuccess:
                Cons_printf("Ok (DMA Int received)\n");
                break;

            case ApiWaitTimeout:
                Cons_printf("*ERROR* - Timeout waiting for Int Event\n");
                break;

            case ApiWaitCanceled:
                Cons_printf("*ERROR* - Interrupt event cancelled\n");
                break;

            default:
                Cons_printf("*ERROR* - API failed\n");
                PlxSdkErrorDisplay(rc);
                break;
        }
    }
    else
    {
        Cons_printf("*ERROR* - API failed\n");
        PlxSdkErrorDisplay(rc);
    }


    Cons_printf("  Check notification status...... ");
    rc =
        PlxPci_NotificationStatus(
            pDevice,
            &NotifyObject,
            &PlxInterrupt
            );

    if (rc == ApiSuccess)
    {
        Cons_printf("Ok (triggered ints:");

        if (PlxInterrupt.DmaChannel_0 == 1)
            Cons_printf(" DMA_0");

        if (PlxInterrupt.DmaChannel_1 == 1)
            Cons_printf(" DMA_1");

        Cons_printf(")\n");
    }
    else
    {
        Cons_printf("*ERROR* - API failed\n");
        PlxSdkErrorDisplay(rc);
    }


    // Transfer a user-mode buffer using DMA
    Cons_printf("  Transfer a user-mode buffer.... ");

    // Allocate a buffer
    pBuffer = malloc(0x10);

    // Clear DMA data
    memset(&DmaParams, 0, sizeof(PLX_DMA_PARAMS));

    DmaParams.u.UserVa      = (PLX_UINT_PTR)pBuffer;
    DmaParams.LocalAddr     = LocalAddress;
    DmaParams.ByteCount     = 0x10;
    DmaParams.LocalToPciDma = 1;

    rc =
        PlxPci_DmaTransferUserBuffer(
            pDevice,
            DmaChannel,
            &DmaParams,
            200         // Specify a timeout to let API perform wait
            );

    if (rc == ApiSuccess)
    {
        Cons_printf("Ok\n");
    }
    else
    {
        Cons_printf("*ERROR* - API failed\n");
        PlxSdkErrorDisplay(rc);
    }


    // Release the interrupt wait object
    Cons_printf("  Cancelling Int Notification.... ");
    rc =
        PlxPci_NotificationCancel(
            pDevice,
            &NotifyObject
            );

    if (rc != ApiSuccess)
    {
        Cons_printf("*ERROR* - API failed\n");
        PlxSdkErrorDisplay(rc);
    }
    else
    {
        Cons_printf("Ok\n");
    }

    // Close DMA Channel
    Cons_printf("  Close DMA Channel.............. ");
    rc =
        PlxPci_DmaChannelClose(
            pDevice,
            DmaChannel
            );

    if (rc == ApiSuccess)
    {
        Cons_printf("Ok\n");
    }
    else
    {
        Cons_printf("*ERROR* - API failed\n");
        PlxSdkErrorDisplay(rc);
    }

    // Release user buffer
    if (pBuffer != NULL)
        free(pBuffer);
}
